# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Validate a json configuration file against a json schema."""

import sys
import re
from datetime import datetime
import logging
import numpy as np
import valideer as v


class InputValidator:

    def __init__(self, json_config, json_schema):
        """Accept a json configuration and a schema to instantiate an InputValidator object."""

        self._json_config = json_config
        self._json_schema = json_schema
        self.validate_input_file()

    @property
    def json_config(self):
        return self._json_config

    @property
    def json_schema(self):
        return self._json_schema

    def validate_input_file(self):
        """Wrapper to validation functions required to validate input file."""

        self.validate_config_structure()
        validate_subset_dict(self._json_schema, self._json_config)
        logging.info("---Successfully validated the file---")

    def validate_config_structure(self):
        """Validate the json file using the validate method of valideer.

        :return: None
        :raises: ValidationError if the validation is not successful
        """
        logging.info("Validating the skeleton of the file...")
        validator = v.parse(self._json_schema)
        try:
            validator.validate(self._json_config)
            logging.info("Successfully validated the skeleton of the file")

        except v.base.ValidationError as e:
            logging.exception("Input validation failed. Please check your input and try again... \n %s", str(e))
            raise


def validate_subset_cols(use_cols=None, subset_cols=None, subset_cols_name='subset_cols'):
    """Accept two lists, use_cols and subset_cols, and issue an error if subset_cols is not a subset of use_cols.

    :param use_cols: the parent list (this is usually the use_cols field in the configuration file)
    :param subset_cols: a subset of the use_cols. This is usually other fields in column section, e.g. fillnan_cols
    :param subset_cols_name: a string that will be used to print an output message
    :return: None
    """
    logging.info("Checking if %s exist in valid columns...", subset_cols_name)
    diff = set(subset_cols) - set(use_cols)
    if diff:
        logging.critical("The columns %s do not exist in valid columns. Exiting...", diff)
        sys.exit(1)


def validate_subset_list(parent_list=None, parent_name=None, subset_list=None, subset_name=None):
    """Check if a list is a subset of another list or not.

    :param parent_list: the parent list (the bigger list)
    :param parent_name: a string that indicates the name of the parent list
    :param subset_list: the subset list (the smaller list)
    :param subset_name: a string that indicates the name of the subset list
    :return: None
    """
    logging.info("Checking if %s exist in %s...", subset_name, parent_name)
    diff = set(subset_list) - set(parent_list)
    if diff:
        logging.critical("Invalid item(s): %s. Exiting...", diff)
        sys.exit(1)


def validate_intersection_cols(cols1=None, cols2=None, cols1_name="cols1", cols2_name="cols2"):
    """Accept two lists and issue an error if there is an intersection between the two.

    :param cols1: a list of columns of a category type
    :param cols2: a list of columns of a category type
    :param cols1_name: a string associated with the name of cols1 (this is used to print an output message)
    :param cols2_name: a string associated with the name of cols2 (this is used to print an output message)
    :return: None
    """
    logging.info("Checking whether %s and %s have an intersection...", cols1_name, cols2_name)
    if cols1 and cols2:
        inter = set(cols1).intersection(set(cols2))

        if inter:
            logging.critical("%s and %s both contain %s. These two must be mutually exclusive. Exiting...",
                          cols1_name, cols2_name, inter)
            sys.exit(1)


def validate_subset_dict(parent_dict=None, subset_dict=None, *args):
    """Accept two dictionaries, parent_dict and subset_dict, and issue an error if subset_dict has keys that
    do not exist in the parent_dict

    :param parent_dict: the parent dictionary that validation will be performed against
    :param subset_dict: the subset dictionary that needs to be validated
    :return: list of strings used to print an output message
    :raise: Error if sub-field(s) exist in config but not in schema due to a typo
    :raise: Warning if default values to be used
    """
    schema_keys = [re.sub('^\+', '', key) for key in parent_dict.keys()]
    diff = set(subset_dict.keys()) - set(schema_keys)

    if diff:
        logging.critical("Input validation failed. Make sure configuration json file matches the schema. "
                         "The error is %s:%s in config file are not contained in schema file. Exiting...",
                         ":".join(list(args)), diff)
        sys.exit(1)

    for key in parent_dict.keys():
        try:
            key_val = re.sub('^\+', '', key)
            if isinstance(parent_dict[key], dict) and parent_dict[key]:
                validate_subset_dict(parent_dict[key], subset_dict[key_val], key_val)
        except KeyError:
            logging.info(f"No parameters for the optional field '{key}' was provided. Learner will use the default "
                         f"values for all subfields.")


def remove_subset_list(parent_list, subset_list):
    """Accepts two lists and returns an updated parent list (without modifying the original). This function does not
    raise an error if the elements of the subset_list do not exist in the parent_list

    :param parent_list: the parent list the removing operation will be performed on
    :param subset_list: a subset_list that should be removed from the parent list if the elements exist in the parent list
    :return: an updated copy of parent list after removing subset_cols
    """
    tmp_list = parent_list[:]
    for col in subset_list:
        if col in tmp_list:
            tmp_list.remove(col)
    return tmp_list


def validate_in_range(numbers, minimum=-np.inf, maximum=np.inf):
    """Check if a number or a list of numbers fall within a range (inclusive). If not, issue an error. This function
    skips None in case it appears in the list, i.e. it does not raise an error for None

    :param numbers: a single number or a list of numbers
    :param minimum: the minimum value for defining the range
    :param maximum: the maximum value for defining the range
    :return: None
    """
    logging.info("Checking if the numbers are between {minimum} and {maximum}".format(minimum=minimum, maximum=maximum))
    # convert the numbers to list
    if not numbers:
        return
    if isinstance(numbers, (int, float)):
        numbers = [numbers]

    for number in numbers:
        # make sure we don't exit on None items
        if number is None:
            continue
        if not minimum <= number <= maximum:
            logging.error("The number {number} is not between {minimum} and {maximum}. Exiting...".
                          format(number=number,
                                 minimum=minimum,
                                 maximum=maximum))
            sys.exit(1)


def validate_type(object_, type_, object_name, type_name):
    """Get an object and validate its type. Provide a message to explain why the validation failed.

    :param object_: the object that we need to check its type
    :param type_: the type we want the object to have
    :param object_name: the name of the object. This will be used for logging purposes
    :param type_name: the name that we want to use for target type. This can be a user-friendly name
    :return: None
    """
    logging.info(f"Checking if the {object_name} is of type {type_name}")

    if not isinstance(object_, type_):
        logging.error(f"The {object_name} does not have the correct type. Expected {type_} got {type(object_)}")
        sys.exit(1)


def validate_date_format(date: str, date_format: str="%Y-%m-%d", field_name: str=None) -> None:
    """Accept a date and a date format to validate the format of the date. If the format of the date is invalid, issue
    an error. An optional field_name can also be provided for a more descriptive log. The field name would hint the
    user what field in the configuration file is being validated.

    :param date: a string that indicates a date
    :param date_format: the format to use for checking the format of the date
    :param field_name: the field name in the configuration file used for logging purposes
    :return: None
    """
    if field_name:
        logging.info(f"Validating the date format in {field_name} to make sure it follows the format: {date_format}")
    else:
        logging.info(f"Validating {date} to make sure it follows the format: {date_format}")

    try:
        datetime.strptime(date, date_format)
    except ValueError:
        logging.critical(f"The date '{date}' has an invalid format. The format should be {date_format}. Exiting...")
        sys.exit(1)



#! /usr/bin/env python

# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

from learner.setup.main import main

if __name__ == '__main__':
    main()



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Manage all the operations related to data. This module communicates with data worker modules to load,
validate, and process the data"""

import logging
from collections.abc import Iterable
import pandas as pd

from learner.data_worker.data_processor import DataProcessor
from learner.data_worker.data_loader import DataLoader, get_data, get_indices_for_value
from learner.data_worker.data_sampler import DataSampler
from learner.data_worker.output_handler import OutputHandler
from learner.feature_engineering.feature_engineering import FeatureEngineering
from learner.validator.data_validator import DataValidator
from learner.outlier_manager.outlier_manager import OutlierManager
from learner.data_worker.data_segmenters import SegmenterHandler
from learner.utilities.timer import timeit
from learner.validator.input_validator import remove_subset_list
from learner.analysis.analysis import Analysis


class TrainDataManager:
    """Handle loading, validating, and processing the train data."""

    def __init__(self, conf):
        """Accept a conf object (from the Configuration class), and handle loading, processing, and validating the
        train data.

        :param conf: a conf object
        """
        self._conf = conf
        self.data = None
        self.processor = None
        self.feature_engineering = None
        self.sample = None

        if self._conf.model.train_models:
            self.get_data()

    @property
    def conf(self):
        return self._conf

    @timeit("load the training data")
    def get_data(self):
        """Load the training data.

        :return: None
        """
        logging.info("Loading the train data...")
        loader = DataLoader(self._conf)
        self.data = loader.load_train_data()
        self.save_data()
        self.validate_data()
        self.analyze_data()
        self.sample_data()
        self.handle_outlier()
        self.process_data()
        self.handle_feature_engineering()
        segmenter_handler = SegmenterHandler(self._conf)
        segmenter_handler.handle_segmenter(self.data)

    @timeit("save the data")
    def save_data(self):
        """If requested by user, save the data obtained using the query into a file.

        :return: None
        """
        if self._conf.data.train_query_activate and self._conf.data.train_query_save_to_file:
            format_ = self._conf.data.train_format
            logging.info(f"Saving the train data obtained from the query into a {format_} file")
            output = OutputHandler(self._conf)
            filename = f"train_data_from_query{self.conf.sep_timetag}.{format_}"
            output.save_file(self.conf.workspace.data_directory,
                             filename,
                             self.data,
                             add_timetag=False,
                             format=format_,
                             sep=self.conf.data.train_delimiter)
            logging.info("Successfully saved the train data into the data directory")

    @timeit("filter the outliers")
    def handle_outlier(self):
        """Instantiate an outlier manager object and make the call to handle_outlier to get the updated data.

        :return: None
        """
        outlier_manager = OutlierManager(self._conf, self.data)
        self.data = outlier_manager.handle_outlier()

    @timeit("validate the training data")
    def validate_data(self):
        """Validate the data

        :return: None
        """
        if self._conf.validation.validate:
            validator = DataValidator(self._conf, self.data, data_type="train")
            validator.validate_data()

    @timeit("analyze the training data")
    def analyze_data(self):
        """Analyze the data

        :return: None
        """
        analysis = Analysis(self._conf)
        analysis.analyze_data(self.data)

    @timeit("sample the training data")
    def sample_data(self):
        """Sample the training data if necessary. Note that we may not need to sample the data in many situations. In
        that case, the driver method (sample_data) would make the decision.

        :return: the updated training dataset
        """
        self.sample = DataSampler(self.conf, self.data)
        self.data = self.sample.sample_data()

    @timeit("process the training data")
    def process_data(self):
        """Process the data. Data processing is carried out based on the user's instructions in the configuration file.

        :return: None
        """
        self.processor = DataProcessor(self._conf)
        self._conf, self.data = self.processor.process_data(self.data)

    def handle_feature_engineering(self):
        self.feature_engineering = FeatureEngineering(self._conf)
        self.data = self.feature_engineering.handle_feature_engineering(self.data)


class ValidationDataManager:
    """Handle loading, validating, and processing the validation data."""

    def __init__(self, conf, processor=None, feature_engineering=None):
        """Accept a conf object (from the Configuration class), and handle loading, processing, and validating the
        train data.

        :param conf: a conf object
        :param processor: a DataProcessor object
        :param feature_engineering: a FeatureEngineering object
        """
        self._conf = conf
        self.data = None
        self.processor = processor
        self.feature_engineering = feature_engineering
        # we need the validation data only if we want to score the models
        self.get_data()

    @property
    def conf(self):
        return self._conf

    @timeit("load the validation data")
    def get_data(self):
        """Load the validation data.

        :return: None
        """
        logging.info("Loading the validation data...")
        loader = DataLoader(self._conf)
        self.data = loader.load_validation_data()
        self.save_data()
        self.validate_data()
        self.handle_outlier()
        self.process_data()
        self.handle_feature_engineering()
        segmenter_handler = SegmenterHandler(self._conf, data_type="validation")
        segmenter_handler.handle_segmenter(self.data)

    @timeit("save the data")
    def save_data(self):
        """If requested by user, save the data obtained using the query into a file.

        :return: None
        """
        if self._conf.data.validation_query_activate and self._conf.data.validation_query_save_to_file:
            format_ = self._conf.data.validation_format
            logging.info(f"Saving the validation data obtained from the query into a {format_} file")
            output = OutputHandler(self._conf)
            filename = f"validation_data_from_query{self.conf.sep_timetag}.{format_}"
            output.save_file(self.conf.workspace.data_directory,
                             filename,
                             self.data,
                             add_timetag=False,
                             format=format_,
                             sep=self.conf.data.train_delimiter)
            logging.info("Successfully saved the validation data into the data directory")

    @timeit("filter the outliers")
    def handle_outlier(self):
        """Instantiate an outlier manager object and make the call to handle_outlier to get the updated data.

        :return: None
        """
        outlier_manager = OutlierManager(self._conf, self.data)
        self.data = outlier_manager.handle_outlier()

    @timeit("validate the validation data")
    def validate_data(self):
        """Validate the data

        :return: None
        """
        if self._conf.validation.validate:
            validator = DataValidator(self._conf, self.data, data_type="validation")
            validator.validate_data()

    @timeit("process the validation data")
    def process_data(self):
        """Process the data. Data processing is carried out based on the user's instructions in the configuration file.

        :return: None
        """
        self._conf, self.data = self.processor.process_data(self.data)

    def handle_feature_engineering(self):
        self.feature_engineering = FeatureEngineering(self._conf)
        self.data = self.feature_engineering.handle_feature_engineering(self.data)


class TestDataManager:
    """Handle loading, validating, and processing the test data."""

    # setting this so that pytest would not try to run this as a test
    __test__ = False

    def __init__(self, conf, processor=None, feature_engineering=None):
        """Accept a conf object (from the Configuration class), and handle loading, processing, and validating the
        test data through iteration.

        :param conf: a conf object
        :param processor: a DataProcessor object
        :param feature_engineering: a FeatureEngineering object
        """
        self._conf = conf
        self._file_nrows = None
        self.processor = processor
        self.feature_engineering = feature_engineering
        self.save_data()

    @property
    def conf(self):
        return self._conf

    @property
    def file_nrows(self):
        return self._file_nrows

    def save_data(self):
        """In some situations, we need to load the entire test data in memory and the save it into disk. These situations
        include using query and segmentation or using query and setting save_to_file to true. In those cases, use
        data loader to get the data and save it to a file.

        :return: None
        """
        # if requested or segmenter's been activated, we need to get all the test data and save it
        if (self._conf.data.test_query_activate and self._conf.data.test_query_save_to_file) or \
           (self._conf.data.test_query_activate and self.conf.segmenter.activate):
            format_ = self._conf.data.test_format
            logging.info(f"Getting the test data using the query and saving it into a {format_} file")
            loader = DataLoader(self._conf)
            data = loader.load_test_from_db()
            data = data[self.conf.column.use_cols] if self.conf.column.use_cols else data
            output = OutputHandler(self._conf)
            filename = f"test_data_from_query{self.conf.sep_timetag}.{format_}"
            output.save_file(self.conf.workspace.data_directory,
                             filename,
                             data,
                             add_timetag=False,
                             format=format_,
                             sep=self.conf.data.test_delimiter)
            logging.info("Successfully saved the test data into the data directory. Learner will use this file for making "
                         "predictions moving forward")
            self.conf.data.test_location = self.conf.workspace.data_directory + filename

    def get_reader(self):
        """Use load_data to get a reader object or a list containing a single dataframe. When using binary file formats
        like parquet or feather, we have to load the entire data into memory because chunksize becomes irrelevant. Since
        the drivers iterate through the test data, we send a list (an iterable) so that the interface remains seamless.

        :return: TextFileReader object or a list
        """
        logging.info("Getting a reader to process the test data...")
        loader = DataLoader(self._conf)
        reader = loader.load_test_data()
        # If reader is an iterable and if it's not a pandas dataframe return it otherwise we will turn it into an
        # iterable (a list in this case) so we can treat all formats the same
        import pandas
        if isinstance(reader, Iterable) and not isinstance(reader, pd.DataFrame):
            return reader
        else:
            return [reader]

    def get_reader_data(self, chunk):
        """Validate, process, and return the processed dataset.

        :param chunk: a pandas dataframe
        :return: a processed pandas dataframe
        """
        # make sure we only keep use_cols if we are not reading from a file
        if not self._conf.data.test_location:
            chunk = chunk[remove_subset_list(self._conf.column.use_cols, [self._conf.column.target_col])]

        self.validate_data(chunk)
        _, chunk = self.process_data(chunk)
        chunk = self.handle_feature_engineering(chunk)
        segmenter_handler = SegmenterHandler(self._conf)
        segmenter_handler.handle_segmenter(chunk)
        return chunk

    def validate_data(self, data):
        """Validate the data.

        :param data: a pandas dataframe
        :return: None
        """
        if self._conf.validation.validate:
            validator = DataValidator(self._conf, data, data_type="test")
            validator.validate_data()

    def process_data(self, data):
        """Process the data. Data processing is carried out based on user input in the config file.

        :param data: a pandas dataframe to be processed
        :return: a processed pandas dataframe
        """
        # we'd probably never need to check this
        if self.processor is None:
            self.processor = DataProcessor(self._conf)
        _, data = self.processor.process_data(data)
        return _, data

    def handle_feature_engineering(self, chunk):
        """Get a chunk and run it through various feature_engineering processes.

        :param chunk: a pandas dataframe loaded in memory
        :return: the updated dataframe after feature engineering
        """
        if self.feature_engineering is None:
            self.feature_engineering = FeatureEngineering(self._conf)
        chunk = self.feature_engineering.handle_feature_engineering(chunk)
        return chunk

    def get_reader_for_segment(self, col, seg_border):
        """Get a TextFileReader (or a list containing a df) and obtain the indices for loading the test data in situations
        where there are multiple segments (the recommender engine currently makes use of this method).

        :param col: column name to process for getting test
        :param seg_border: the segment border
        :return: a TextFileReader (or a list containing a df) and a numpy array containing the indices in the entire dataset for seg id
        """
        logging.info("Getting a reader to process the test data...")

        data = get_data(self._conf.data.test_location,
                        manifest_file=self._conf.data.manifest,
                        format=self._conf.data.test_format,
                        sep=self._conf.data.test_delimiter,
                        nrows=self._conf.data.test_nrows,
                        usecols=[col],
                        header=self._conf.data.test_header)

        segmenter_handler = SegmenterHandler(self._conf)
        segmenter_handler.handle_segmenter(data)

        indices = get_indices_for_value(data, self._conf.segmenter.test_name, seg_border)

        logging.info(f"There are {len(indices)} rows of data in this segment")

        loader = DataLoader(self._conf)
        loader.connect_file(self._conf.data.test_location,
                            chunksize=self._conf.data.test_chunksize,
                            nrows=self._conf.data.test_nrows,
                            usecols=remove_subset_list(self._conf.column.use_cols, [self._conf.column.target_col]),
                            format=self._conf.data.test_format)

        if self._conf.data.test_format == "csv":
            return loader.df, indices
        else:
            return [loader.df], indices




# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""A parser module responsible for parsing command line arguments and other parsing tasks as they emerge."""

import sys
import os
import ntpath


def get_filename_from_path(path):
    """Extract the filename from the filepath.

    :param path: the filepath
    :return: return the file name (after the final slash).
    """
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)


def get_config_file():  # pragma: no cover
    """Get the path to the configuration file. If nothing is provided, use a sample configuration file from the config
     folder

    :return: the path to a configuration file
    """
    try:
        config_path = sys.argv[1]
        config_file = os.path.abspath(config_path)
        current_path = os.path.dirname(os.path.realpath(__file__))
        parent_path = os.path.abspath(os.path.join(current_path, os.pardir))
        os.chdir(parent_path)
    except IndexError:
        config_file = "config/classifier.json"
    config_filename = get_filename_from_path(config_file)
    if os.path.exists(config_file):
        return config_file, config_filename
    return config_file, config_filename


# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import os
from datetime import datetime

from learner.setup.parser import get_config_file
from learner.setup.setup import Setup


class EngineHandler:
    """A handler class to instantiate and run appropriate engines depending on user input and other settings."""

    def __init__(self, conf):
        self._conf = conf

        self.handle_engine()

    def handle_engine(self):
        if self._conf.engine == "Recommender" and self._conf.segmenter.activate:
            from learner.engines.recommender import RecommenderSegment
            engine = RecommenderSegment(self._conf)
        elif self._conf.engine == "Recommender":
            from learner.engines.recommender import Recommender
            engine = Recommender(self._conf)
        elif self._conf.segmenter.activate:
            from learner.engines.standard_engines import StandardEngineSegment
            engine = StandardEngineSegment(self._conf)
        elif self._conf.engine in ("Classifier", "Regressor"):
            from learner.engines.standard_engines import StandardEngine
            engine = StandardEngine(self._conf)
        elif self._conf.engine == "DeepClassifier":
            from learner.engines.deep_classifier import DeepClassifier
            engine = DeepClassifier(self._conf)
        elif self._conf.engine == "DeepRegressor":
            from learner.engines.deep_regressor import DeepRegressor
            engine = DeepRegressor(self._conf)
        elif self._conf.engine == "ImageClassifier":
            from learner.engines.image_classifier import ImageClassifier
            engine = ImageClassifier(self._conf)
        engine.run_engine()



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements the ColumnValidator class to validate the column names defined in different section of the
configuration file.
"""
import re
import warnings
import logging

from learner.validator.input_validator import validate_subset_cols, validate_intersection_cols


class ColumnValidator:
    def __init__(self, conf):
        self.conf = conf

    def validate_columns(self):
        """The main method that communicates with other methods to validate the columns defined in different sections
        of the configuration file. The name of each validator method ends with the section name.

        :return: None
        """
        self.validate_analysis()
        self.validate_column()
        self.validate_data()
        self.validate_feature_engineering()
        self.validate_outlier()
        self.validate_process()
        self.validate_sample()

    def validate_analysis(self):
        validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=self.conf.analysis.correlation_cols,
                             subset_cols_name="correlation columns")

        validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=self.conf.analysis.shap_exclude_cols,
                             subset_cols_name="shap exclude columns")

        # this is the pattern for finding the values between curly braces
        pattern = re.compile("{(.*?)}")
        if self.conf.analysis.shap_narrative_stories:
            for class_name, stories in self.conf.analysis.shap_narrative_stories.items():
                for item in stories["templates"]:
                    # validate the columns in the "features" section
                    validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=item["features"].keys(),
                                         subset_cols_name="features columns")

                    # now validate the template columns to make sure they exist in use columns
                    results = re.findall(pattern, item["narrative"])
                    validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=results,
                                         subset_cols_name="columns in narrative text")

    def validate_column(self):
        validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=self.conf.column.copy_cols.keys(),
                             subset_cols_name="copy_cols")

        if self.conf.column.dtype_activate and self.conf.column.col_dtypes:
            validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=self.conf.column.col_dtypes.keys(),
                                 subset_cols_name="cols_dtypes")

        validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=[self.conf.column.target_col], subset_cols_name="target_col")

        # one can request the items of date column to show up in the prediction output
        validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=self.conf.column.id_cols, subset_cols_name="id_cols")

        validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=self.conf.column.drop_cols, subset_cols_name="drop_cols")

        validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=self.conf.column.drop_from_train, subset_cols_name="drop_from_train")

    def validate_data(self):
        # make sure join_cols are included in id_cols
        diff = set(self.conf.data.validation_join_cols) - set(self.conf.column.id_cols)
        if diff and self.conf.data.validation_score_activate:
            self.conf.column.id_cols = set(self.conf.column.id_cols).union(diff)
            self.conf.column.id_cols = list(self.conf.column.id_cols)

            warnings.warn(f"The columns {diff} were included in scoring_params join_cols but not column id_cols. "
                          "Updating the column id_cols to include this...", UserWarning)

        # make sure if segmenter is activated we have join_cols defined
        if not self.conf.data.validation_join_cols and self.conf.segmenter.activate and self.conf.data.validation_score_activate:
            warnings.warn("Segmentation's been activated but join_cols are not defined in the scoring_params "
                          "section. Learner will try to use id_cols defined in column section. This could fail if "
                          "those columns have been processed or changed after loading them", UserWarning)
            self.conf.data.validation_join_cols = self.conf.column.id_cols

        validate_subset_cols(use_cols=self.conf.column.valid_cols, subset_cols=self.conf.data.validation_join_cols,
                             subset_cols_name="scoring join_cols")

    def validate_feature_engineering(self):
        if self.conf.feature_engineering.basic_operations_params:
            for item in self.conf.feature_engineering.basic_operations_params:
                # validate only if the item is activated
                if item["activate"]:
                    validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                         subset_cols=item["cols"],
                                         subset_cols_name="basic_operations_cols")

        if self.conf.feature_engineering.log_transform_params:
            for item in self.conf.feature_engineering.log_transform_params:
                # validate only if the item is activated
                if item["activate"]:
                    validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                         subset_cols=[item["col"]],
                                         subset_cols_name="log transform cols")

        if self.conf.feature_engineering.groupby_params:
            for param in self.conf.feature_engineering.groupby_params:
                # validate only if the item is activated
                if param["activate"]:
                    validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                         subset_cols=[param["col"]],
                                         subset_cols_name="groupby cols")
                    for agg in param["aggregation"]:
                        validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                             subset_cols=[agg["col"]],
                                             subset_cols_name="groupby aggregation cols")

    def validate_outlier(self):
        validate_subset_cols(use_cols=self.conf.column.valid_cols,
                             subset_cols=self.conf.outlier.min_max_dict,
                             subset_cols_name="min_max_params")

        validate_subset_cols(use_cols=self.conf.column.valid_cols,
                             subset_cols=self.conf.outlier.quantile_dict,
                             subset_cols_name="quantile_params")

        validate_subset_cols(use_cols=self.conf.column.valid_cols,
                             subset_cols=self.conf.outlier.sd_dict,
                             subset_cols_name="sd_params")

        validate_subset_cols(use_cols=self.conf.column.valid_cols,
                             subset_cols=self.conf.outlier.value_cols.keys(),
                             subset_cols_name="cols_params in value_params")

    def validate_process(self):
        if self.conf.process.fillnan_activate and self.conf.process.fillnan_value_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.fillnan_value_cols.keys(),
                                 subset_cols_name="fillnan cols")

        if self.conf.process.fillnan_activate and self.conf.process.fillnan_mean_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.fillnan_mean_cols,
                                 subset_cols_name="fillnan mean cols")

        if self.conf.process.fillnan_activate and self.conf.process.fillnan_median_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.fillnan_median_cols,
                                 subset_cols_name="fillnan median cols")

        if self.conf.process.fillnan_activate and self.conf.process.fillnan_mode_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.fillnan_mode_cols,
                                 subset_cols_name="fillnan mode cols")

        if self.conf.process.dummies_activate and self.conf.process.dummies_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.dummies_cols,
                                 subset_cols_name="dummies cols")
            # we don't want to process categorical columns for deep learning engines
            if self.conf.process.dummies_activate is True and self.conf.engine in ("DeepClassifier", "DeepRegressor"):
                self.conf.process.dummies_activate = False

        if self.conf.process.to_numeric_activate and self.conf.process.to_numeric_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.to_numeric_cols,
                                 subset_cols_name="to_numeric cols")

        if self.conf.process.label_encoding_activate and self.conf.process.label_encoding_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.label_encoding_cols,
                                 subset_cols_name="label_encoding cols")

        if self.conf.process.tuplize_activate and self.conf.process.tuplize_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.tuplize_cols,
                                 subset_cols_name="tuplize cols")

        if self.conf.process.date_cols_activate and self.conf.process.date_cols_params:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.date_cols_params.keys(),
                                 subset_cols_name="date cols")

        if self.conf.process.standard_scaler_activate and self.conf.process.standard_scaler_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.standard_scaler_cols,
                                 subset_cols_name="standard scaler cols")

        if self.conf.process.min_max_scaler_activate and self.conf.process.min_max_scaler_cols:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=self.conf.process.min_max_scaler_cols,
                                 subset_cols_name="min_max scaler cols")

        count_vectorize_cols = []
        if self.conf.process.count_vectorize_activate:
            for col_param in self.conf.process.count_vectorize_cols_params:
                count_vectorize_cols.append(col_param["name"])
                validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                     subset_cols=[col_param["name"]],
                                     subset_cols_name="count vectorize cols")

        tfidf_cols = []
        if self.conf.process.tfidf_activate:
            for col_param in self.conf.process.tfidf_cols_params:
                tfidf_cols.append(col_param["name"])
                validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                     subset_cols=[col_param["name"]],
                                     subset_cols_name="tfidf cols")

        # make sure count_vectorize and tfidf_cols don't have an intersection
        if self.conf.process.count_vectorize_activate and self.conf.process.tfidf_activate:
            validate_intersection_cols(cols1=count_vectorize_cols,
                                       cols1_name="count_vectorize cols",
                                       cols2=tfidf_cols,
                                       cols2_name="tfidf cols")

    def validate_sample(self):
        if self.conf.sample.split_activate and self.conf.sample.split_method == "sort" and self.conf.sample.split_sort_col:
            validate_subset_cols(use_cols=self.conf.column.valid_cols,
                                 subset_cols=[self.conf.sample.split_sort_col],
                                 subset_cols_name="col in options of train_test_split_params")



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import logging
import os
import sys

from learner.data_worker.data_loader import get_value


class WorkspaceConfiguration:
    """Parse all input variables related to the workspace section including the name of the files and the method
    for saving the files."""

    def __init__(self, json_config, output_path, verbose_level, output_name):
        self._json_config = json_config

        self.path = output_path
        self.verbose_level = verbose_level
        self.name = output_name
        self.concat_output = get_value(self._json_config, True, "workspace", "concat_output")
        self.data_directory = self.get_data_directory()
        self.s3_activate = get_value(self._json_config, False, "workspace", "s3_params", "activate")
        self.s3_path = self.get_s3_path()

    def get_data_directory(self):  # pragma: no cover
        # create a directory called "data" under output.path to save the split data there. Currently, this is the first
        # place that needs to handle this. This may change in the future.
        data_directory = self.path + "data/"
        if not os.path.exists(data_directory):
            os.makedirs(data_directory)
        return data_directory

    def get_s3_path(self):
        try:
            path = self._json_config["workspace"]["s3_params"]["path"]
            if self.s3_activate and not path.startswith("s3://"):
                logging.critical("s3 was activated but 'path' does not start with 's3://'. Please update your "
                                 "configuration file and try again. Exiting...")
                sys.exit(1)
            return path
        except KeyError:
            if self.s3_activate:
                logging.critical("s3_params was activated but 'path' is not defined. Please update your configuration "
                                 "file and try again. Exiting...")
                sys.exit(1)
            return None



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements the ImageClassifier engine."""
import os
import logging
import pandas as pd
import torch

from learner.utilities.timer import timeit
from learner.configuration.configuration import Configuration
from learner.data_worker.image_data_manager import TrainDataManager, ValidationDataManager, TestDataManager
from learner.model_manager.image_classifiers import ImageClassifierHandler
from learner.model_manager.model_initializer import get_pickle_dir
from learner.model_manager.prediction_manager import ImageClassifierPredictor
from learner.data_worker.output_handler import OutputHandler


class ImageClassifier:
    def __init__(self, conf: Configuration):
        """Take a conf object (from the Configuration class) to run a image_classifier engine.

        :param conf: a conf object
        """
        self._conf = conf
        # we are not using model_initializer here but I do this to keep things consistent with other engines
        self.models_dict = self.conf.model.models_dict
        self.conf.model.nrows_score = 0

    @property
    def conf(self):
        return self._conf

    def run_engine(self):
        """The main method that communicates with other methods to drive the engine.

        :return: None
        """
        if self._conf.model.train_models:
            self.train_models()
            self.predict()
        else:
            logging.info("Loading saved model...")
            self.load_models(self._conf.workspace.path)

    @timeit("train the model(s)")
    def train_models(self):
        """Loop through all the items in models_dict and call the train_models method to train each model. Note that,
        unlike other engines, saving the model happens during the training.

        :return: None
        """
        logging.info("Building the models...")
        for tag, mdl in self.conf.model.models_dict.items():
            logging.info(f"Fitting {tag} model")
            self.fit_models(tag, mdl)

    def fit_models(self, tag, mdl):
        """Get the train and validation (if needed) data loaders. Then, call the appropriate methods to train the
        models. Currently, we only fit the models but we'll probably implement learning rate finder soon.

        :param tag: the tag of the model (arbitrary tag defined by the user)
        :param mdl: an items in models_dict
        :return: None
        """
        train_data_manager = TrainDataManager(self.conf, mdl)
        train_loader = train_data_manager.get_data_loader()
        validation_loader = None
        if self._conf.data.validation_score_activate:
            validation_data_manager = ValidationDataManager(self.conf, mdl)
            validation_loader = validation_data_manager.get_data_loader()

        if self._conf.model.lr_activate:
            logging.info("Finding the learning rate...")
            mdl["model"] = ImageClassifierHandler.learner_find_learning_rate(tag=tag,
                                                                             mdl=mdl,
                                                                             conf=self.conf,
                                                                             train_loader=train_loader)
        else:
            mdl["model"] = ImageClassifierHandler.learner_fit(tag=tag,
                                                              mdl=mdl,
                                                              conf=self.conf,
                                                              train_loader=train_loader,
                                                              validation_loader=validation_loader)

    @timeit("load the model(s)")
    def load_models(self, output_path):
        """Load saved models from the provided path if full directory is provided. Otherwise, find and load the latest
        saved models.

        :param output_path: directory path to the saved models
        :return: None
        """
        _, pickled_file_names, _ = get_pickle_dir(output_path, ext="pth")
        for file_name in pickled_file_names:
            checkpoint = torch.load(file_name)
            self.models_dict = {checkpoint["tag"]: checkpoint["mdl"]}
            checkpoint["conf"].sep_timetag = self.conf.sep_timetag
            # we want to use the current data parameters
            checkpoint["conf"].data = self._conf.data
            # we want to use the current combine parameters
            checkpoint["conf"].combine = self._conf.combine
            self._conf = checkpoint["conf"]
            self.predict()

    @timeit("make predictions using the trained model(s)")
    def predict(self):
        """Call predict_test if we need to make predictions on test data. Unlike other engines, scoring the validation
        data happens during training.

        :return: None
        """
        if self._conf.data.test_prediction_activate:
            logging.info("Making predictions using test data...")
            self.predict_test()

    def predict_test(self):
        """Use the trained models (stored in models_dict) to make predictions. Making predictions are performed by
        iterating through the test dataset. As it happens, there are a lot of similarities between image_classifier and
        other non-deep learning engines.

        :return: None
        """
        test_data_manager = TestDataManager(self.conf, mdl=None)
        test_loader = test_data_manager.get_data_loader()

        # we probably don't need to keep track of the num_chunks because we can get it from the length of data loader
        # but I think this is fine.
        num_chunks = 0
        for index, (X, data) in enumerate(test_loader):
            # data will be a dictionary ready to be converted to a dataframe
            data = pd.DataFrame(data=data)
            num_chunks += 1

            for tag, mdl in self.models_dict.items():
                predictor = ImageClassifierPredictor(mdl, self._conf)
                prediction_df, data = predictor.make_predictions(X,
                                                                 data,
                                                                 data_type="test",
                                                                 column_name=self._conf.data.test_column_name,
                                                                 prediction_type=self._conf.data.test_prediction_type)

                predictor.write_predictions(tag,
                                            prediction_df,
                                            data,
                                            index,
                                            data_type="test",
                                            add_timetag=self._conf.data.test_add_timetag,
                                            add_version=self._conf.model.version_activate,
                                            version_name=self._conf.model.version_version_name,
                                            version_column_name=self._conf.model.version_column_name
                                            )

            # keep track of number of rows in test dataset for validation purposes
            self.conf.model.nrows_score += X.shape[0]

        output = OutputHandler(self._conf, data_type="test")
        output.concat_chunk_csv(num_chunks, self.models_dict)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements all scoring functions that are not readily available"""

import math
from sklearn.metrics import mean_squared_error


def root_mean_squared_error(*args, **kwargs):
    """Because root_mean_squared_error is not readily available but mean_squared_error is, we compute mean_squared_error
    and return the squared root to get root_mean_squared_error.

    :param args: the positional arguments, these are y_true and y_pred
    :param kwargs: the keywords arguments if there are any
    :return: the root mean squared error
    """
    return math.sqrt(mean_squared_error(args[0], args[1], **kwargs))



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

from learner.data_worker.data_loader import get_value


class RecommenderConfiguration:
    """Parse all input variables related to recommender engine including the minimum and maximum length for a basket
    of a product."""

    def __init__(self, json_config):
        self._json_config = json_config
        self.support_cutoff = get_value(self._json_config, 0.2, "recommender", "support_cutoff")
        self.min_length = get_value(self._json_config, 1, "recommender", "min_length")
        self.max_length = get_value(self._json_config, 10, "recommender", "max_length")
        self.max_num_category = get_value(self._json_config, 100, "recommender", "max_num_category")
        self.activate = get_value(self._json_config, True, "recommender", "extend_params", "activate")
        self.num_recs = get_value(self._json_config, 5, "recommender", "extend_params", "num_recs")




# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import warnings


class ValidationConfiguration:

    def __init__(self, json_config):
        self._json_config = json_config

        self.validate = self.get_validate()

    def get_validate(self):
        try:
            validate = self._json_config["validation"]["validate"]
            if not validate:
                warnings.warn("Validation flag is set to false, this can lead to errors or unexpected results",
                              UserWarning)
            return validate
        except KeyError:
            return True



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Manage email communications including authentication, preparing and sending messages."""
import os
import smtplib
import logging
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email import encoders

from learner.configuration.configuration import Configuration


class BaseEmailManager:
    """The parent class for email manager subclasses."""
    def __init__(self, conf=None, message=None, data=None):
        self._conf = conf
        self._data = data
        self._message = message

    @property
    def conf(self):
        return self._conf

    @property
    def data(self):
        return self._data

    @property
    def message(self):
        return self._message

    @staticmethod
    def send_email(smtp, msg, username, email_list):
        """Use an authenticated smtp object and a message to send emails.

        :param smtp: an smtp object
        :param msg: the message object to be sent
        :param username: the email username
        :param email_list: the list of emails to send to
        :return: None
        """
        logging.info("Sending emails...")
        smtp.sendmail(username, email_list, msg.as_string())
        smtp.quit()
        logging.info("Emails were sent successfully...")

    def authenticate(self, username, password):
        """Authenticate to the mail server

        :param username: the email username
        :param password: the email password
        :return: an authenticated smtp object
        """
        logging.info("Authenticating to the mail server...")
        smtp = smtplib.SMTP(self.conf.communication.admin_email_host, self.conf.communication.admin_email_port)
        smtp.starttls()
        smtp.login(username, password)
        logging.info("Successfully authenticated to the mail server...")
        return smtp

    @staticmethod
    def prepare_email_content(body, attachment_paths=None):
        """Prepare the email content with body in text. append attachment if provided.

        :param body: the text to use for the body of the email
        :param attachment_paths: list of paths to the attachment file
        :return: an email object
        :rtype: MIMEMultipart
        """
        email = MIMEMultipart()
        text = MIMEText(body, "html")
        email.attach(text)

        if attachment_paths is not None:
            for attachment_path in attachment_paths:
                p = MIMEBase('application', 'octet-stream')
                p.set_payload(open(attachment_path, "rb").read())
                encoders.encode_base64(p)
                p.add_header('Content-Disposition', "attachment; filename= {path}".format(path=os.path.split(attachment_path)[-1]))
                email.attach(p)

        return email


class AdminEmailManager(BaseEmailManager):
    """Email managing system for administrative purposes."""
    def __init__(self, conf: Configuration, message=None, data=None):
        super(AdminEmailManager, self).__init__(conf, message=message, data=data)
        self.smtp = None

    def handle_communications(self):
        """Call appropriate methods to make decisions and prepare messages for different communication methods.

        :return: None
        """
        self.smtp = self.authenticate(self._conf.communication.admin_email_ausername, self._conf.communication.admin_email_password)
        msg = getattr(self, "prepare_admin_message")()
        self.send_email(self.smtp, msg, self._conf.communication.admin_email_ausername, self._conf.communication.admin_email_address)

    def prepare_admin_message(self):
        """Prepare the admin email including body, subject, etc.

        :return: an email object
        """
        logging.info("Preparing admin email message...")
        body = self.message

        email = self.prepare_email_content(body)
        email['Subject'] = "[LEARNER ALERT]: " + body
        email['From'] = self._conf.communication.admin_email_ausername
        email['To'] = ", ".join(self._conf.communication.admin_email_address)

        logging.info("Successfully prepared the email message...")
        return email



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import logging
import warnings
from datetime import datetime
from learner.configuration.supported_items import SUPPORTED_IMBALANCED_METHODS, SUPPORTED_SPLIT_METHODS
from learner.data_worker.data_loader import get_value
from learner.configuration.workspace import WorkspaceConfiguration
from learner.validator.input_validator import validate_date_format


class SampleConfiguration:
    def __init__(self, json_config, engine, workspace: WorkspaceConfiguration):
        self._json_config = json_config
        self._engine = engine
        self._workspace = workspace

        self.split_data = self.get_split_data()
        self.imbalanced_activate = self.get_imbalanced_activate()
        self.imbalanced_method = self.get_imbalanced_method()
        self.imbalanced_random_state = get_value(self._json_config, 42, "sample", "imbalanced_params", "random_state")

        self.split_activate = get_value(self._json_config, False, "sample", "train_test_split_params", "activate")
        self.split_method = self.get_split_method()
        self.split_save_train_data = get_value(self._json_config, False, "sample", "train_test_split_params", "save_train_data")

        self.split_test_size = self.get_split_test_size()

        self.split_random_state = get_value(self._json_config, None, "sample", "train_test_split_params", "random_state")
        self.split_shuffle = get_value(self._json_config, True, "sample", "train_test_split_params", "shuffle")
        self.split_random_stratify = get_value(self._json_config, None, "sample", "train_test_split_params", "stratify")

        self.split_sort_col = get_value(self._json_config, None, "sample", "train_test_split_params", "col")
        self.split_sort_nan_position = self.get_split_sort_nan_position()
        self.split_test_on_after = self.get_split_test_on_after()
        self.split_train_before = self.get_split_train_before()

    @property
    def json_config(self):  # pragma: no cover
        return self._json_config

    @property
    def engine(self):
        return self._engine

    @property
    def output(self):
        return self._workspace

    def get_split_data(self):
        if self.engine in ("DeepClassifier", "DeepRegressor"):
            return False
        return True

    def get_imbalanced_activate(self):
        try:
            activate = self._json_config["sample"]["imbalanced_params"]["activate"]
            acceptable_engines = ("Classifier",)
            if activate and self.engine not in acceptable_engines:
                warnings.warn("imbalanced_params were activated but the engine is not {acceptable_engines}. "
                              "Deactivating the imbalanced_params".format(acceptable_engines=acceptable_engines),
                              UserWarning)
                return False
            return activate
        except KeyError:
            return False

    def get_imbalanced_method(self):
        try:
            method = self._json_config["sample"]["imbalanced_params"]["method"].lower()
            if self.imbalanced_activate and method not in SUPPORTED_IMBALANCED_METHODS.values():
                try:
                    return SUPPORTED_IMBALANCED_METHODS[method[0]]
                except KeyError:
                    logging.error("Invalid method in imbalanced_params, supported methods are {supported_methods}"
                                  .format(supported_methods=SUPPORTED_IMBALANCED_METHODS.values()))
                    sys.exit(1)
            return method
        except KeyError:
            return "undersampling"

    def get_split_method(self):
        try:
            method = self._json_config["sample"]["train_test_split_params"]["method"].lower()
            if self.split_activate and method not in SUPPORTED_SPLIT_METHODS.values():
                try:
                    return SUPPORTED_SPLIT_METHODS[method[0]]
                except KeyError:
                    logging.critical("""Invalid method in train_test_split_params, supported methods are 
                                     {supported_methods}""".format(supported_methods=SUPPORTED_SPLIT_METHODS.values()))
                    sys.exit(1)
            return method
        except KeyError:
            return "random"

    def get_split_sort_nan_position(self):
        supported_nan_position = {"f": "first", "l": "last"}
        try:
            nan_position = self._json_config["sample"]["train_test_split_params"]["nan_position"].lower()
            if self.split_activate and self.split_method == "sort" and supported_nan_position.values():
                try:
                    return supported_nan_position[nan_position[0]]
                except KeyError:
                    logging.critical(f"""Invalid value nan_position of train_test_split_params, accepted values are  
                                     {supported_nan_position.values()}""")
        except KeyError:
            return "last"

    def get_split_test_size(self):
        try:
            test_size = self._json_config["sample"]["train_test_split_params"]["test_size"]
            if self.split_activate:
                if test_size < 0:
                    logging.critical("The test_size in train_test_split_params cannot be negative")
                    sys.exit(1)
            # when the test size is larger than one, that means the user has defined the number of samples they want
            # not the fraction of data
            if test_size > 1:
                return int(test_size)
            return test_size
        except KeyError:
            return 0.2

    def get_split_test_on_after(self):
        date_format = "%Y-%m-%d"
        try:
            test_after = self._json_config["sample"]["train_test_split_params"]["test_on_after"]
            if self.split_activate and self.split_method == "sort":
                validate_date_format(date=test_after, date_format=date_format,
                                     field_name="test_on_after")
            test_after = datetime.strptime(test_after, date_format)
            return test_after
        except KeyError:
            return None

    def get_split_train_before(self):
        date_format = "%Y-%m-%d"
        try:
            train_before = self._json_config["sample"]["train_test_split_params"]["train_before"]
            if self.split_activate and self.split_method == "sort":
                validate_date_format(date=train_before, date_format=date_format,
                                     field_name="train_before")
            train_before = datetime.strptime(train_before, date_format)
            if not self.split_test_on_after:
                self.split_test_on_after = train_before
            return train_before
        except KeyError:
            return self.split_test_on_after




# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import functools

from learner.validator.input_validator import remove_subset_list


def exclude(name=None):
    """Accept a named argument and update that argument by deleting the items in the exclude_list and then run the
    function.

    :param name: a named argument in a function that will be decorated
    :return: a decorator function
    """
    def exclude_decorator(method):
        @functools.wraps(wrapped=method)
        def exclude_values(*args, **kwargs):
            try:
                kwargs[name] = remove_subset_list(kwargs[name], kwargs["exclude_list"])
            except KeyError:
                pass
            result = method(*args, **kwargs)
            return result
        return exclude_values
    return exclude_decorator



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module responsible for reading and loading the datasets."""

import sys
import random
import json
import warnings
import logging
import pandas as pd

from learner.validator.input_validator import remove_subset_list


class DataLoader:
    def __init__(self, conf=None):
        """The conf is an instance of the Configuration class. The df attribute can either hold a panda's dataframe
        once it's being loaded in memory or it can be a TextFileReader object."""

        self._conf = conf
        self.df = None
        self.meta_data = None
        self.train_nrows = self._conf.data.train_nrows
        self.validation_nrows = self._conf.data.validation_nrows
        self.skiprows = None

    @property
    def conf(self):
        return self._conf

    def connect_file(self, data_source, nrows=None, chunksize=None, dtype=None, usecols=None, sep=",", header=0, call_num=0, format="csv"):
        """Load data through a file location. The df attribute gets populated here. This method accepts a file format and
        then calls the relevant methods to load the data.

        :param data_source: the path to the data file
        :param nrows: number of rows of the file to read
        :param chunksize: the chunksize when iterating through the file, setting it will return a TextFileReader object
        :param dtype: a dictionary containing the dtypes
        :param usecols: a list of columns that should be used when loading the data
        :param call_num: call number for the attempts to load the data. This is used to avoid infinite recursion
        :param format: the format of the file - the supported formats are: csv, parquet, and feather
        :return: None
        """
        if format == "csv":
            self.connect_csv(data_source, nrows, chunksize, dtype, usecols, sep, header, call_num)
        elif format == "parquet":
            self.connect_parquet(data_source, usecols)
        elif format == "feather":
            self.connect_feather(data_source, usecols)
        # we should never get here as long as configuration file is validated
        else:
            logging.critical("Unknown file format. Exiting...")
            sys.exit(1)

    def connect_csv(self, data_source, nrows=None, chunksize=None, dtype=None, usecols=None, sep=",", header=0, call_num=0):
        """Load data from a csv file. If the dtype is populated, we'll try to load the data using those dtypes. If we
        fail to lead the data, we'll ignore the dtype and make a second attempt.

        :param data_source: the path to the data file
        :param nrows: number of rows of the file to read
        :param chunksize: the chunksize when iterating through the file, setting it will return a TextFileReader object
        :param dtype: a dictionary containing the dtypes
        :param usecols: a list of columns that should be used when loading the data
        :param call_num: call number for the attempts to load the data. This is used to avoid infinite recursion
        :return: None
        """
        try:
            # try to load with dtypes if it's being asked
            self.df = pd.read_csv(data_source
                                  , sep=sep
                                  , names=self._conf.column.all_cols
                                  , nrows=nrows
                                  , header=header
                                  , usecols=usecols
                                  , skiprows=self.skiprows
                                  , chunksize=chunksize
                                  , dtype=dtype)

            logging.info("Successfully loaded the data")
        except Exception as e:
            if call_num > 0:
                logging.critical("The second attempt for loading data was unsuccessful. Exiting...")
                sys.exit(1)
            logging.error("Loading the data failed, trying to load the data with updated parameters. The error is: '%s'",
                          str(e))
            # increment call_num to avoid infinite recursion (call_num+1)
            self.connect_csv(data_source, nrows=nrows, usecols=usecols, sep=sep, header=header, call_num=call_num + 1)

    def connect_parquet(self, data_source, usecols=None, engine="pyarrow", call_num=0):
        """Load the data from a parquet file.

        :param data_source: the path to the data file
        :param usecols: a list of columns that should be used when loading the data
        :return: None
        """
        try:
            self.df = pd.read_parquet(data_source,
                                      columns=usecols,
                                      engine=engine)

            logging.info("Successfully loaded the data")
        except Exception as e:
            if call_num > 0:
                logging.critical("The second attempt for loading data was unsuccessful. Exiting...")
                sys.exit(1)
            logging.error(f"Loading the data failed, trying to load the data with updated parameters. The error is: {str(e)}")
            self.connect_parquet(data_source, usecols=usecols, engine="fastparquet", call_num=call_num + 1)  # we change the engine if we fail

    def connect_feather(self, data_source, usecols=None):
        """Load the data from a feather file.

        :param data_source: the path to the data file
        :param usecols: a list of columns that should be used when loading the data
        :return: None
        """
        try:
            self.df = pd.read_feather(data_source,
                                      columns=usecols)

            logging.info("Successfully loaded the data")
        except Exception as e:
            logging.error(f"Loading the data failed. The error is: {str(e)}")
            sys.exit(1)

    def connect_presto(self, query, chunksize=None):
        """Execute a query and load the results into a pandas dataframe from a presto/hive table

        :param query: a string of query to be executed.
        :param chunksize: if specified, return an iterator where chunksize is the number of rows to include in each chunk
        :return: a pandas dataframe or a generator
        """
        try:
            with self.conf.connection.presto_client.connect() as con:
                self.df = pd.read_sql(query, con=con, chunksize=chunksize)
        except Exception as e:
            logging.critical(f"Loading the data from presto failed. The error is: {str(e)}")
            sys.exit(1)

    def connect_postgres(self, query, chunksize=None):
        """Execute a query and load the results into a pandas dataframe from a postgres table

        :param query: a string of query to be executed.
        :param chunksize: if specified, return an iterator where chunksize is the number of rows to include in each chunk
        :return: a pandas dataframe or a generator
        """
        try:
            with self.conf.connection.postgres_client.connect() as con:
                self.df = pd.read_sql(query, con=con, chunksize=chunksize)
        except Exception as e:
            logging.critical(f"Loading the data from postgres failed. The error is: {str(e)}")
            sys.exit(1)

    def connect_mysql(self, query, chunksize=None):
        """Execute a query and load the results into a pandas dataframe from a mysql table

        :param query: a string of query to be executed.
        :param chunksize: if specified, return an iterator where chunksize is the number of rows to include in each chunk
        :return: a pandas dataframe or a generator
        """
        try:
            with self.conf.connection.mysql_client.connect() as con:
                self.df = pd.read_sql(query, con=con, chunksize=chunksize)
        except Exception as e:
            logging.critical(f"Loading the data from mysql failed. The error is: {str(e)}")
            sys.exit(1)

    def connect_snowflake(self, query, chunksize=None):
        """Execute a query and load the results into a pandas dataframe from a snowflake table

        :param query: a string of query to be executed.
        :param chunksize: if specified, return an iterator where chunksize is the number of rows to include in each chunk
        :return: a pandas dataframe or a generator
        """
        try:
            with self.conf.connection.snowflake_client.connect() as con:
                self.df = pd.read_sql(query, con=con, chunksize=chunksize)
        except Exception as e:
            logging.critical(f"Loading the data from snowflake failed. The error is: {str(e)}")
            sys.exit(1)

    def load_train_data(self):
        """Load the training data by calling the appropriate methods.

        :return: pandas dataframe
        """
        logging.info("Loading the training data...")
        if self._conf.data.train_location:
            return self.load_train_from_file()
        elif self._conf.data.train_query_activate:
            return self.load_train_from_db()

    def load_validation_data(self):
        """Load the validation data by calling the appropriate methods.

        :return: pandas dataframe
        """
        logging.info("Loading the validation data...")
        if self._conf.data.validation_location:
            return self.load_validation_from_file()
        elif self._conf.data.validation_query_activate:
            return self.load_validation_from_db()

    def load_test_data(self):
        """Get a pointer to the test data so that we can iterate through the chunks.

        :return: a generator
        """
        logging.info("Loading the test data...")
        if self._conf.data.test_location:
            return self.load_test_from_file()
        elif self._conf.data.test_query_activate:
            return self.load_test_from_db(chunksize=self._conf.data.test_chunksize)

    def load_train_from_file(self):
        """If we need to load training data from a file, decide how to load the data and call connect_file to get them.

        :return: a pandas dataframe containing the training data
        """
        if self._conf.column.dtype_activate:
            logging.info("Loading the data using dtypes...")
            # update the dtype dictionary by doing a trial run
            self._conf.column.dtype_dict = self.update_dtype_dict(self._conf.data.train_location,
                                                                  dtype_trial_nrows=self._conf.column.dtype_trial_nrows)

        if self._conf.data.train_sample_size and self._conf.data.train_format == "csv":
            logging.info("Sampling %i rows from the train dataset...", self._conf.data.train_sample_size)
            self.train_nrows = self.get_file_num_rows(self._conf.data.train_location)

            if self._conf.data.train_sample_size >= self.train_nrows:
                warnings.warn("The sample size is greater than the number of rows in the dataset. The entire "
                              "dataset will be used for training...", Warning)
            else:
                self.skiprows = self.get_skiprows(self.train_nrows)

        self.connect_file(self._conf.data.train_location,
                          nrows=self.train_nrows,
                          dtype=self._conf.column.dtype_dict,
                          usecols=self._conf.column.use_cols,
                          sep=self._conf.data.train_delimiter,
                          header=self._conf.data.train_header,
                          format=self._conf.data.train_format
                          )

        # reset indices if we've skipped rows
        if self.skiprows:
            self.df.reset_index(drop=True, inplace=True)

        if self._conf.data.train_sample_size and self._conf.data.train_format != "csv":
            logging.info("Sampling %i rows from the train dataset...", self._conf.data.train_sample_size)
            self.df = self.df.sample(n=self._conf.data.train_sample_size,
                                     replace=False,
                                     random_state=self._conf.data.train_sample_seed)
            # the newer version of pandas has the option "ignore_index", we can update when we use the newer versions
            self.df.reset_index(drop=True, inplace=True)

        return self.df

    def load_validation_from_file(self):
        """If we need to load validation data from a file, decide how to load the data and call connect_file to get them.

        :return: a pandas dataframe containing the validation data
        """
        if self._conf.column.dtype_activate:
            logging.info("Loading the data using dtypes...")
            # update the dtype dictionary by doing a trial run
            self._conf.column.dtype_dict = self.update_dtype_dict(self._conf.data.validation_location,
                                                                  dtype_trial_nrows=self._conf.column.dtype_trial_nrows)

        if self._conf.data.validation_sample_size:
            logging.info("Sampling %i rows from the validation dataset...", self._conf.data.validation_sample_size)
            self.validation_nrows = self.get_file_num_rows(self._conf.data.validation_location)

            if self._conf.data.validation_sample_size >= self.validation_nrows:
                warnings.warn("The sample size is greater than the number of rows in the dataset. The entire "
                              "dataset will be used for training...", Warning)
            else:
                self.skiprows = self.get_skiprows(self.validation_nrows)

        self.connect_file(self._conf.data.validation_location,
                          nrows=self.validation_nrows,
                          dtype=self._conf.column.dtype_dict,
                          usecols=self._conf.column.use_cols,
                          sep=self._conf.data.validation_delimiter,
                          header=self._conf.data.validation_header,
                          format=self._conf.data.validation_format
                          )

        # reset indices if we've skipped rows
        if self.skiprows:
            self.df.reset_index(drop=True, inplace=True)

        return self.df

    def load_test_from_file(self, **kwargs):
        """If we need to get the test data from a file, call connect_file with chunksize to get a TextFileReader object.

        :return: a TextFileReader object from pandas
        """
        self.connect_file(self._conf.data.test_location,
                          nrows=self._conf.data.test_nrows,
                          chunksize=kwargs["chunksize"] if "chunksize" in kwargs else self._conf.data.test_chunksize,
                          usecols=remove_subset_list(self._conf.column.use_cols, [self._conf.column.target_col]),
                          sep=self._conf.data.test_delimiter,
                          header=self._conf.data.test_header,
                          format=self._conf.data.test_format)

        return self.df

    def load_train_from_db(self):
        """If we need to load the training data from a table, check what type of db we need to connect to and then call
        the appropriate method to ge the data. Once we obtain the data, make sure we only keep the columns defined in use_cols.

        :return: a pandas dataframe
        """
        if self._conf.data.train_db_type == "presto":
            self.connect_presto(self._conf.data.train_query)
        elif self._conf.data.train_db_type == "postgres":
            self.connect_postgres(self._conf.data.train_query)
        elif self._conf.data.train_db_type == "mysql":
            self.connect_mysql(self._conf.data.train_query)
        elif self._conf.data.train_db_type == "snowflake":
            self.connect_snowflake(self._conf.data.train_query)

        if self._conf.column.use_cols:
            # only keep the columns in use columns
            self.df = self.df[self._conf.column.use_cols]

        return self.df

    def load_validation_from_db(self):
        """If we need to load the validation data from a table, check what type of db we need to connect to and then call
        the appropriate method to ge the data. Once we obtain the data, make sure we only keep the columns defined in use_cols.

        :return: a pandas dataframe
        """
        if self._conf.data.validation_db_type == "presto":
            self.connect_presto(self._conf.data.validation_query)
        elif self._conf.data.validation_db_type == "postgres":
            self.connect_postgres(self._conf.data.validation_query)
        elif self._conf.data.validation_db_type == "mysql":
            self.connect_mysql(self._conf.data.validation_query)
        elif self._conf.data.validation_db_type == "snowflake":
            self.connect_snowflake(self._conf.data.validation_query)

        if self._conf.column.use_cols:
            # only keep the columns in use columns
            self.df = self.df[self._conf.column.use_cols]

        return self.df

    def load_test_from_db(self, chunksize=None):
        """If we need to load the test data from a table, check what type of db we need to connect to and then call
        the appropriate method to ge the data. The chunksize can be passed as a parameter here. The reason is that, this
        method can be used to get a TextFileReader object or the actual data. We may need to get the actual data, if the
        user needs to save the query data or wants to use query along with segmentation.

        :param chunksize: the chunksize for getting the data
        :return: A TextFileReader object or the loaded pandas dataframe.
        """

        if self._conf.data.test_db_type == "presto":
            self.connect_presto(self._conf.data.test_query, chunksize=chunksize)
        elif self._conf.data.test_db_type == "postgres":
            self.connect_postgres(self._conf.data.test_query, chunksize=chunksize)
        elif self._conf.data.test_db_type == "mysql":
            self.connect_mysql(self._conf.data.test_query, chunksize=chunksize)
        elif self._conf.data.test_db_type == "snowflake":
            self.connect_snowflake(self._conf.data.test_query, chunksize=chunksize)

        return self.df

    def update_dtype_dict(self, data_source, dtype_trial_nrows=200000):
        """Update dtype_dict through a trial load.

        :param data_source: path to the data
        :param dtype_trial_nrows: the number of rows to be used for the trial run
        :return: an updated dtype_dict
        """
        # do trial loading of data to remove cols with incorrect dtypes
        logging.info("Checking if dtype_dict is valid. dtype_dict will be updated for invalid types...")
        for col in self._conf.column.use_cols:
            if col in self._conf.column.dtype_dict:
                try:
                    pd.read_csv(data_source,
                                sep=self._conf.data.delimiter,
                                names=self._conf.column.all_cols,
                                nrows=dtype_trial_nrows,
                                header=self._conf.data.header,
                                usecols=[col],
                                dtype=self._conf.column.dtype_dict[col])
                except Exception as e:
                    del self._conf.column.dtype_dict[col]
                    logging.info("Deleting %s from dtype_dict due to incorrect type. The error is: '%s'", col, str(e))

        return self._conf.column.dtype_dict

    def get_file_num_rows(self, location):
        """Get the number of rows in a dataset by reading one column from the file.

        :return: number of rows in a dataset
        """
        logging.info("Getting the total number of rows in the dataset...")
        file = pd.read_csv(location,
                           delimiter=self._conf.data.train_delimiter,
                           usecols=[0],
                           header=self._conf.data.train_header,
                           skip_blank_lines=False)

        # add 1 to file_num_rows if the dataset doesn't have a header row because we are assuming one of the rows does
        # not contain data
        if self._conf.data.manifest:
            file_num_rows = file.shape[0] + 1
        else:
            file_num_rows = file.shape[0]

        logging.info("The dataset has %i rows", file_num_rows)
        return file_num_rows

    def get_skiprows(self, file_num_rows):
        """Use the number of rows in a data set and the sample size to return the rows numbers that should be skipped
        when reading a data set. This value is then passed to the pandas reader.

        :return: a list of indices that should be skipped
        """
        # randomly select the row numbers that should NOT be loaded (we do this because panda provide the functionality
        # for it). We sample (file_num_rows - sample_size) numbers from the total number of rows in the file.
        # for example if there are 100 rows in the data set and we want to sample 30 rows. We randomly select 70 rows
        # that should be skipped
        random.seed(a=self._conf.data.train_sample_seed)
        return sorted(random.sample(range(self._conf.data.train_header + 1, file_num_rows),
                                    file_num_rows - self._conf.data.train_sample_size))


def get_data(data_source, manifest_file=None, format="csv", **kwargs):
    """Load a dataset and return a pandas dataframe. This function supports multiple file formats including csv, parquet,
    and feather.

    :param data_source: the path to the dataset
    :param manifest_file: the path to a manifest file if it exists
    :param format: the format of the file. The supported formats are: csv, parquet, and feather
    :param kwargs: the kwargs that should be passed to read_csv
    :return: a pandas dataframe
    """
    if manifest_file:
        idx, cols = get_index_cols_from_manifest(manifest_file, kwargs["usecols"])
        kwargs['names'] = cols
        kwargs['usecols'] = idx
        kwargs['header'] = None

    if format == "csv":
        data = pd.read_csv(data_source, **kwargs)
    elif format == "parquet":
        data = pd.read_parquet(data_source, columns=kwargs["usecols"])
        # if we have the manifest, we plugin the names here
        if "names" in kwargs:
            data.columns = kwargs["names"]
    elif format == "feather":
        data = pd.read_feather(data_source, columns=kwargs["usecols"])
        # if we have the manifest, we plugin the names here
        if "names" in kwargs:
            data.columns = kwargs["names"]
    else:
        logging.critical("Unknown data type. Exiting...")
        sys.exit(1)
    return data


def get_indices_for_value(data, col, value):
    """Get the indices of a dataframe corresponding to a value in a specific column.

    :param data: a pandas dataframe
    :param col: a column name to perform the filtering on. This should be a string
    :param value: the value to be used for filtering the data set
    :return: a numpy array that contains the indices corresponding to value
    :raises TypeError: if an incorrect type for col is passes, e.g, list instead of string
    """
    if isinstance(col, str) is False:
        logging.error("get_indices_for_value accept a string, not %s", type(col))
        raise TypeError()
    indices = data[data[col] == value].index.values
    return indices


def get_index_cols_from_manifest(manifest_file, usecols):
    """Get the indices and the corresponding column names using a manifest file. This function returns the correct order
    of column names if the usecols passed by user does not match the sequence of the column names in manifest.
    This function assumes the manifest file contain the correct sequence for the column names.

    :param manifest_file: a manifest file that contains column names. Each row in this file contains the name of a single column
    :param usecols: a list of column names, which should be used to get the indices
    :return: a list of indices and a list of column names in the right order
    """
    with open(manifest_file, 'r') as file:
        cols = [x.strip() for x in file.readlines()]
        if usecols:
            idx = [cols.index(c) for c in cols if c in usecols]
        else:
            idx = [cols.index(c) for c in cols]
    # loop through the indices and retrieve a list of column names
    selected_cols = [cols[i] for i in idx]
    return idx, selected_cols


def get_meta_data(meta_data_path):
    """Get the meta data loaded from the meta file listed in the config

    :param meta_data_path: path of json or csv file to the meta data file to open
    :return: a dictionary containing meta data
    """
    try:
        with open(meta_data_path) as f:
            meta_data = json.load(f)
        return meta_data
    except json.decoder.JSONDecodeError:
        logging.exception("Decoding meta data json file has failed. Check the formatting of the json file")
        raise


def get_file_num_rows(location, **kwargs):
    """Get the number of rows in a dataset by reading one column from the file.
    Note: this function only works for datasets that contain a header row. This function is typically used for
    validating the output files generated by Learner.

    :param location: location of file to retrieve number of rows
    :param kwargs: parameters used by pandas read_csv function
    :return: number of rows in a dataset
    """
    file = pd.read_csv(location,
                       usecols=[0],
                       skip_blank_lines=False,
                       **kwargs)

    file_num_rows = file.shape[0]

    return file_num_rows


def get_dtype_dict(meta_data, use_cols, col_dtypes):
    """Get column dtypes from either meta_data or configuration file (priority)

    :param meta_data: a meta_data file
    :param use_cols: the use_cols for loading the data
    :param col_dtypes: the col_dtype dictionary from the configuration file
    :return: the updated dtype_dict
    """
    # initialize dtype_dict from configuration file and meta data file
    dtype_dict = {}
    if meta_data:
        for col in meta_data['column']:
            if col in use_cols:
                try:
                    dtype_dict[col] = meta_data['column'][col]['dtype']
                except KeyError:
                    logging.info("The col %s exists in meta_data file but doesn't include dtype...", col)

    if col_dtypes:
        for col in col_dtypes:
            dtype_dict[col] = col_dtypes[col]
    return dtype_dict


def get_value(dictionary, default_value, *keys):
    """Get the value from a single- or multi-level (nested) dictionary using a series of keys. If the value does not
    exist, return the default_value.

    :param dictionary: a single- or multi-level dictionary
    :param default_value: return this value if no value exist
    :param keys: a chain of keys pointing towards the value needed
    """
    tmp = dictionary
    for key in keys:
        if isinstance(tmp, dict):
            try:
                tmp = tmp[key]
            except KeyError:
                return default_value
    return tmp



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for validating the output files generated by Learner."""

import sys
import logging

from learner.data_worker.data_loader import get_file_num_rows


class OutputValidator:
    def __init__(self, conf, models_dict):
        """Validates the outputs of prediction process.

         :param conf: configuration object
         :param models_dict: dict containing the model tag and number of rows scored
         """
        self._conf = conf
        self._models_dict = models_dict

    @property
    def conf(self):
        return self._conf

    @property
    def models_dict(self):
        return self._models_dict

    def validate(self):
        """The main function that runs all the instance methods to validate the outputs generated by Learner if the
        validation flag is set to true.

        :return: None
        """
        logging.info("Validating the output...")
        for tag, mdls in self._models_dict.items():
            mdl = mdls.get(0, mdls)
            self.validate_num_predictions(tag, mdl)
        logging.info("Successfully validated the output")

    def validate_num_predictions(self, tag, mdl):
        """Check if the output file generated by Learner contain the expected number of rows or not.

        :param tag: The tag for the model
        :return: None
        :raises SystemExit: if the number of rows in the file doesn't match the expected value
        """
        logging.info("Validating number of predictions")
        filename = mdl["path"] + self._conf.workspace.name + "_test_" + str(tag) + str(self._conf.sep_timetag) + ".csv"
        file_num_rows = get_file_num_rows(filename)

        diff = self.conf.model.nrows_score - file_num_rows
        if diff != 0:
            logging.error("Prediction input and output shapes do not match.\n "
                          "Expected: %i Found: %i", self.conf.model.nrows_score, file_num_rows)
            sys.exit(1)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module to analyze the data as well as the models. This module commonly talks to the plot module to
visualize the results of the analysis
"""
import pandas as pd
import numpy as np
import warnings
import logging

from learner.configuration.configuration import Configuration
from learner.data_worker.output_handler import OutputHandler
from learner.analysis.plot import Plot
from learner.utilities.templates import PREDICTIONS_VS_ACTUALS_PLOT
from learner.model_manager.scoring_manager import RegressorScorer


class Analysis:
    """Implement the methods for analyzing the data and the models. Some of the methods here such as handle_feature_importance
    are being called by the engines."""
    def __init__(self, conf: Configuration):
        """Instantiate an analysis object using the conf object"""
        self._conf = conf

    @property
    def conf(self):
        return self._conf

    def handle_feature_importance(self, mdl, filename):
        """Accept an item in the  models_dict, use the trained model object there to get and save the feature importance.
        If a model does not have the feature_importance_ attribute, issue a warning and return. This method also uses
        the Plot class to plot the results if requested by user.

        :param mdl: an item in the models_dict
        :param filename: the full path to the file that should be used to save the data and plot. This parameter should not
        include the file extension.
        :return: None
        """
        logging.info("Getting feature importance and saving it to the disk...")
        try:
            importances_df = pd.DataFrame(data={"feature": mdl["train_features"],
                                                "importance": mdl["model"].feature_importances_})\
                .sort_values('importance', ascending=False).reset_index(drop=True)

            output = OutputHandler(self._conf)
            full_name = filename + ".csv"
            output.save_file(mdl["path"], full_name, importances_df, add_timetag=False)
            logging.info("The feature importance data was saved in {full_name}".format(full_name=full_name))
        except AttributeError:
            warnings.warn("Feature importance was requested for {model_type} but this model does not provide "
                            "feature importance. Trying to continue...".format(model_type=mdl["type"]), UserWarning)
            return

        if self.conf.analysis.importance_plot_activate:
            logging.info("Plotting the feature importance...")
            # get the top n features
            importances_df = importances_df.iloc[:self.conf.analysis.importance_plot_num_features, :]
            nfeatures = np.arange(importances_df.shape[0])
            plot = Plot(self.conf)
            plot.barh(importances_df["importance"],
                      importances_df["feature"],
                      mdl["path"] + filename+".png",
                      yrange=nfeatures,
                      color=self.conf.analysis.importance_plot_color,
                      width=self.conf.analysis.importance_plot_width or 12,
                      height=self.conf.analysis.importance_plot_height or importances_df.shape[0]/2,
                      xlabel="Feature Importance")

    def handle_missing_values(self, data):
        """Accept a pandas dataframe and compute the count and the ratio of the missing values for all columns in the
        dataframe. Then save the data to disk in csv format and if requested, plot the data using a horizontal bar
        chart.

        :param data: a pandas dataframe
        :return: None
        """
        logging.info("Getting missing values and saving it to the disk...")
        try:
            # get the count of missing values
            missing_df = data.isnull().sum(axis=0).reset_index()
            missing_df.columns = ["feature", "missing_count"]
            missing_df['missing_ratio'] = (missing_df["missing_count"] / data.shape[0])
            missing_df = missing_df.sort_values(by='missing_ratio', ascending=False)
            # save data to disk
            output = OutputHandler(self._conf)
            output.save_file(self.conf.analysis.analysis_directory,
                             self.conf.analysis.missing_filename,
                             missing_df, add_timetag=False)
            logging.info("The missing values data was saved in {filename}".format(filename=self.conf.analysis.missing_filename))
        except Exception as e:
            logging.error("Something went wrong while obtaining the missing values. The error is {error}".format(error=str(e)))
            return

        if self.conf.analysis.missing_plot_activate:
            logging.info("Plotting the missing value data...")
            # get the top n features
            missing_df = missing_df.iloc[:self.conf.analysis.missing_plot_num_features, :]
            nfeatures = np.arange(missing_df.shape[0])
            plot = Plot(self.conf)
            plot.barh(missing_df["missing_ratio"],
                      missing_df["feature"],
                      self.conf.analysis.analysis_directory + self.conf.analysis.missing_plotname,
                      yrange=nfeatures,
                      color=self.conf.analysis.missing_plot_color,
                      width=self.conf.analysis.missing_plot_width or 12,
                      height=self.conf.analysis.missing_plot_height or missing_df.shape[0]/2,
                      xlabel="Missing Ratio")

    def handle_correlation(self, data):
        """Accept a pandas dataframe and compute the correlation coefficient between the features defined in the
        configuration file. Then save the data to disk in csv format and if requested, plot the data using heatmap.

        :param data: a pandas dataframe
        :return: None
        """
        logging.info("Computing the correlation coefficients and saving them to disk...")
        try:
            # compute pearson correlation (the default method when calling corr) coefficient.
            corr = data[self.conf.analysis.correlation_cols].corr()

            # build the mask matrix so that there won't be any duplicates in the plot. We still save everything to disk.
            mask = np.zeros_like(corr)
            mask[np.triu_indices_from(mask, 1)] = True

            # save data to disk
            output = OutputHandler(self._conf)
            output.save_file(self.conf.analysis.analysis_directory,
                             self.conf.analysis.correlation_filename,
                             corr, add_timetag=False, index=True)
            logging.info("The correlation coefficients data was saved in {filename}".format(filename=self.conf.analysis.correlation_filename))
        except Exception as e:
            logging.error("Something went wrong while calculating the correlation coefficients. The error is {error}".format(error=str(e)))
            return

        if self.conf.analysis.correlation_plot_activate:
            logging.info("Plotting the correlation coefficients data...")

            plot = Plot(self.conf)
            plot.correlation_heatmap(corr,
                                     filename=self.conf.analysis.analysis_directory + self.conf.analysis.correlation_plotname,
                                     width=self.conf.analysis.correlation_plot_width or corr.shape[0]/2,
                                     height=self.conf.analysis.correlation_plot_height or corr.shape[0]/2,
                                     x_labelsize=self.conf.analysis.correlation_plot_x_labelsize or corr.shape[0]/2,
                                     y_labelsize=self.conf.analysis.correlation_plot_y_labelsize or corr.shape[0]/2,
                                     annot_size=self.conf.analysis.correlation_plot_annotation_size or corr.shape[0]/2,
                                     vmin=self.conf.analysis.correlation_plot_vmin,
                                     vmax=self.conf.analysis.correlation_plot_vmax,
                                     mask=mask)

    def handle_predictions_vs_actuals_plot(self, tag, mdl):
        """Accept an item of the models_dict (tag and mdl) to handle the predictions vs actuals plot. The mdl dictionary
        and the conf object contain all the necessary information. Because this plot is very identical to scoring the
        predictions (we need y_true and y_pred to do something with them), this method leverages the RegressorScorer
        class to load the necessary data.

        :param tag: a tag (key) in the models_dict item
        :param mdl: a mdl dictionary (the value for the tag) the models_dict
        :return: None
        """
        logging.info("Plotting the predictions vs actuals graph...")
        # the regressor_scorer object will load the actuals and true values for us and apply the log transformation
        # logic if needed
        regressor_scorer = RegressorScorer(self._conf, None)

        pred_filename = mdl["path"] + self._conf.workspace.name + "_validation_" + str(tag) + str(self._conf.sep_timetag) + ".csv"
        regressor_scorer.load_pred(pred_filename)

        plot_filename = PREDICTIONS_VS_ACTUALS_PLOT.format(
            path=mdl["path"],
            output_name=self._conf.workspace.name,
            tag=str(tag),
            sep_timetag=str(self._conf.sep_timetag))

        plot = Plot(self._conf)
        plot.predictions_vs_actuals(regressor_scorer.y_true,
                                    regressor_scorer.pred,
                                    filename=plot_filename,
                                    width=self.conf.analysis.predictions_vs_actuals_width,
                                    height=self.conf.analysis.predictions_vs_actuals_height,
                                    x_label=self.conf.analysis.predictions_vs_actuals_x_label or regressor_scorer.y_true.columns[0],
                                    y_label=self.conf.analysis.predictions_vs_actuals_y_label or regressor_scorer.pred.columns[0],
                                    x_labelsize=self.conf.analysis.predictions_vs_actuals_x_labelsize,
                                    y_labelsize=self.conf.analysis.predictions_vs_actuals_y_labelsize,
                                    color=self.conf.analysis.predictions_vs_actuals_symbol_color,
                                    size=self.conf.analysis.predictions_vs_actuals_symbol_size,
                                    xy_min=self.conf.analysis.predictions_vs_actuals_xy_min,
                                    xy_max=self.conf.analysis.predictions_vs_actuals_xy_max)

    def analyze_data(self, data):
        """The main driver method for calling other methods and running the analysis on data. The method is typically
        called by data_manage module before processing the data. Some methods in this class need to be called after
        model training. As such, this method won't call them.

        :param data: a pandas dataframe (this is usually training data)
        :return: None
        """
        if self.conf.analysis.missing_activate:
            self.handle_missing_values(data)

        if self.conf.analysis.correlation_activate:
            self.handle_correlation(data)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

from learner.data_worker.data_loader import get_value


class SimilaritiesConfiguration:
    """Parse all input variables related to creating a similarity matrix from the input data. Learner provides a wrapper
    around scipy distance package. Therefore, all methods supported by scipy is supported here as well."""

    def __init__(self, json_config):
        self._json_config = json_config
        self.metric = get_value(self._json_config, "jaccard", "similarities", "metric")



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import warnings
import logging

from learner.validator.input_validator import (validate_intersection_cols,
                                               validate_subset_list,
                                               validate_in_range,
                                               validate_type)
from learner.data_worker.data_loader import get_value


class OutlierConfiguration:
    """Parse all inputs related to outlier management of the training data."""

    supported_methods = {"r": "remove", "c": "clip"}

    def __init__(self, json_config, column, data, process):
        self._column = column
        self._json_config = json_config
        self._data = data
        self._process = process

        self.min_max_activate = get_value(self._json_config, False, "outlier", "min_max_params", "activate")
        self.min_max_cols = self.get_min_max_cols()
        self.min_max_dict = self.get_min_max_dict()

        self.quantile_activate = get_value(self._json_config, False, "outlier", "quantile_params", "activate")
        self.quantile_cols = self.get_quantile_cols()
        self.quantile_dict = self.get_quantile_dict()

        self.sd_activate = get_value(self._json_config, False, "outlier", "sd_params", "activate")
        self.sd_cols = self.get_sd_cols()
        self.sd_dict = self.get_sd_dict()

        self.value_activate = get_value(self._json_config, False, "outlier", "value_params", "activate")
        self.value_cols = self.get_value_cols()
        self.value_dict = self.get_value_dict()

    def get_min_max_cols(self):
        try:
            min_max_cols = self._json_config["outlier"]["min_max_params"]["cols_params"]
            # in case min or max does not contain in the keys, add them and set their values to None
            try:
                for col, col_info in min_max_cols.items():
                    # we need to check this because valideer won't
                    validate_subset_list(["min", "max", "method"],
                                         "supported fields",
                                         col_info.keys(),
                                         "the fields in cols_params")

                    min_max_cols[col] = {"min": col_info.get("min", None),
                                         "max": col_info.get("max", None),
                                         "method": col_info.get("method", "remove")}

            except Exception as e:
                logging.error("There is an issue with cols_params in min_max_params. The error is {error}".format(error=str(e)))
                sys.exit(1)
            return min_max_cols
        except KeyError:
            return {}

    def get_min_max_dict(self):
        min_max_dict = {}

        if self._data.meta_data:
            min_max_dict = {col: {"min": col_info["min_max"].get("min", None),
                                  "max": col_info["min_max"].get("max", None),
                                  "method": col_info["min_max"].get("method", "remove")}
                            for col, col_info in self._data.meta_data["column"].items()
                            if col in self._column.use_cols and "min_max" in col_info.keys() and
                            ("min" in col_info["min_max"].keys() or "max" in col_info["min_max"].keys())}

        min_max_dict.update(self.min_max_cols)

        if self.min_max_activate:
            min_max_dict = self.update_min_max_dict(min_max_dict)
            self.validate_min_max_filtering_cols(min_max_dict)
        return min_max_dict

    @staticmethod
    def update_min_max_dict(min_max_dict):
        # make sure we don't keep the items in which both min and max values are None, i.e, no filtering needed
        return {col: {"min": col_info["min"], "max": col_info["max"], "method": col_info["method"]} for col, col_info in min_max_dict.items()
                if (col_info["min"] is not None or col_info["max"] is not None)}

    def validate_min_max_filtering_cols(self, min_max_dict):
        # make sure the specified methods are valid
        for col, col_info in min_max_dict.items():
            try:
                method = col_info["method"]
                col_info["method"] = OutlierConfiguration.supported_methods[method[0].lower()]
            except KeyError:
                logging.error('The method "{method}" for min_max not understood. The acceptable '
                              'methods are {acceptable_methods}'.
                    format(method=method,
                           acceptable_methods=list(OutlierConfiguration.supported_methods.values())))
                sys.exit(1)

        if self._process.dummies_activate:
            validate_intersection_cols(cols1=self._process.dummies_cols,
                                       cols2=list(min_max_dict.keys()),
                                       cols1_name='dummy_cols',
                                       cols2_name='min_max_cols')

        if self._process.label_encoding_activate:
            validate_intersection_cols(cols1=self._process.label_encoding_cols,
                                       cols2=list(min_max_dict.keys()),
                                       cols1_name='label_encoding cols',
                                       cols2_name='min_max_cols')

    def get_quantile_cols(self):
        try:
            quantile_cols = self._json_config["outlier"]["quantile_params"]["cols_params"]
            # in case min or max does not contain in the keys, add them and set their values to None
            try:
                for col, col_info in quantile_cols.items():
                    # we need to check this because valideer won't
                    validate_subset_list(["min", "max", "method"],
                                         "supported fields",
                                         col_info.keys(),
                                         "the fields in cols_params")

                    quantile_cols[col] = {"min": col_info.get("min", None),
                                          "max": col_info.get("max", None),
                                          "method": col_info.get("method", "remove")}

            except Exception as e:
                logging.error("There is an issue with cols_params in quantile_params. The error is {error}".format(error=str(e)))
                sys.exit(1)
            return quantile_cols
        except KeyError:
            return {}

    def get_quantile_dict(self):
        quantile_dict = {}

        if self._data.meta_data:
            quantile_dict = {col: {"min": col_info["quantile"].get("min", None),
                                   "max": col_info["quantile"].get("max", None),
                                   "method": col_info["quantile"].get("method", "remove")}
                             for col, col_info in self._data.meta_data["column"].items()
                             if col in self._column.use_cols and "quantile" in col_info.keys() and
                             ("min" in col_info["quantile"].keys() or "max" in col_info["quantile"].keys())}

        quantile_dict.update(self.quantile_cols)

        if self.quantile_activate:
            quantile_dict = self.update_min_max_dict(quantile_dict)
            self.validate_quantile_cols(quantile_dict)
        return quantile_dict

    @staticmethod
    def update_quantile_dict(quantile_dict):
        # make sure we don't keep the items in which both min and max values are None, i.e, no filtering needed
        return {col: {"min": col_info["min"], "max": col_info["max"], "method": col_info["method"]} for col, col_info in quantile_dict.items()
                if (col_info["min"] is not None or col_info["max"] is not None)}

    def validate_quantile_cols(self, quantile_dict):
        # make sure the specified methods are valid
        for col, col_info in quantile_dict.items():
            try:
                method = col_info["method"]
                col_info["method"] = OutlierConfiguration.supported_methods[method[0].lower()]
                # need to make sure minimum and maximum are between 0 and 1
                validate_in_range([col_info["min"], col_info["max"]], minimum=0, maximum=1)
            except KeyError:
                logging.error('The method "{method}" for quantile not understood. The acceptable '
                              'methods are {acceptable_methods}'
                              .format(method=method,
                                      acceptable_methods=list(OutlierConfiguration.supported_methods.values())))
                sys.exit(1)

        if self._process.dummies_activate:
            validate_intersection_cols(cols1=self._process.dummies_cols,
                                       cols2=list(quantile_dict.keys()),
                                       cols1_name='dummy_cols',
                                       cols2_name='quantile_cols')

        if self._process.label_encoding_activate:
            validate_intersection_cols(cols1=self._process.label_encoding_cols,
                                       cols2=list(quantile_dict.keys()),
                                       cols1_name='label_encoding cols',
                                       cols2_name='quantile_cols')

    def get_sd_cols(self):
        try:
            sd_filtering_cols = self._json_config["outlier"]["sd_params"]["cols_params"]
            return sd_filtering_cols
        except KeyError:
            return {}

    def get_sd_dict(self):
        sd_dict = {}

        if self._data.meta_data:
            sd_dict = dict((col, col_info["sd"]) for col, col_info in self._data.meta_data["column"].items()
                           if col in self._column.use_cols and "sd" in col_info.keys())

        sd_dict.update(self.sd_cols)

        if self.sd_activate:
            self.validate_sd_dict_format(sd_dict)
            self.update_sd_dict(sd_dict)
            self.validate_sd_dict_cols(sd_dict)
        return sd_dict

    @staticmethod
    def validate_sd_dict_format(sd_dict):
        for col, sd_params in sd_dict.items():
            if isinstance(sd_params, dict):
                try:
                    if not isinstance(sd_params["value"], (float, int)):
                        logging.error("The value for col %s in sd_params must be a number. Exiting...",
                                      col)
                        sys.exit(1)
                except KeyError:
                    logging.exception('When passing a dictionary to a column in cols_params, "value" must be '
                                      'defined. A single number can also be passed to a column instead of a '
                                      'dictionary. Exiting...')
                    sys.exit(1)
            elif isinstance(sd_params, (float, int)):
                sd_dict[col] = {'value': sd_params}
            else:
                logging.error("The value for col %s in sd_params must be a number or a dictionary. "
                              "Exiting...", col)
                sys.exit(1)
            # set the method to "remove" if not defined
            sd_dict[col]["method"] = sd_dict[col].get("method", "remove")

    @staticmethod
    def update_sd_dict(sd_dict):
        for col, sd_params in sd_dict.items():
            try:
                if sd_params["center"].lower().startswith('med'):
                    sd_params["center"] = "median"
                elif sd_params["center"].lower().startswith('mo'):
                    sd_params["center"] = "mode"
                else:
                    sd_params["center"] = "mean"
                sd_params["method"] = sd_params.get("method", "remove")
            except KeyError:
                warnings.warn('No center was defined for col {0} in sd_params, using "mean"'.format(col),
                              UserWarning)
                sd_params["center"] = "mean"
            except AttributeError:
                warnings.warn('The defined center for col {0} in sd_params not understood, using "mean"'
                              .format(col), UserWarning)
                sd_params["center"] = "mean"

    def validate_sd_dict_cols(self, sd_dict):
        # make sure the specified methods are valid
        for col, sd_params in sd_dict.items():
            try:
                method = sd_params["method"]
                sd_params["method"] = OutlierConfiguration.supported_methods[method[0].lower()]
            except KeyError:
                logging.error('The method "{method}" for sd_params not understood. The acceptable '
                              'methods are {acceptable_methods}'.
                              format(method=method,
                                     acceptable_methods=list(OutlierConfiguration.supported_methods.values())))
                sys.exit(1)
            # we need to check this because valideer won't
            validate_subset_list(["value", "center", "method"],
                                 "supported fields",
                                 sd_params.keys(),
                                 "the fields in cols_params")

        if self._process.dummies_activate:
            validate_intersection_cols(cols1=self._process.dummies_cols,
                                       cols2=list(sd_dict.keys()),
                                       cols1_name='dummy_cols',
                                       cols2_name='sd_cols')

        if self._process.label_encoding_activate:
            validate_intersection_cols(cols1=self._process.label_encoding_cols,
                                       cols2=list(sd_dict.keys()),
                                       cols1_name='label_encoding cols',
                                       cols2_name='sd_cols')

    def get_value_cols(self):
        try:
            cols_params = self._json_config["outlier"]["value_params"]["cols_params"]
            return cols_params
        except KeyError:
            return {}

    def get_value_dict(self):
        value_dict = {}
        for col, col_info in self.value_cols.items():
            # if col_info is a list, assume all the values need to be remove. So update the dictionary accordingly
            if isinstance(col_info, list):
                value_dict[col] = {}
                for item in col_info:
                    value_dict[col][item] = None
            else:
                # if col_info is not a list, it must be a dictionary
                validate_type(object_=col_info, type_=dict, object_name=f"the value of {col}", type_name="dictionary")
                value_dict[col] = col_info

        return value_dict



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements the DeepClassifier engine. Please note that the majority of the functionality is implemented
in the base class"""

import logging

from learner.configuration.configuration import Configuration
from learner.utilities.timer import timeit
from learner.model_manager.scoring_manager import ClassifierScorer
from learner.combine.combining_manager import MeanClassifierCombiner
from learner.engines.base_deep_engine import BaseDeep


class DeepClassifier(BaseDeep):
    def __init__(self, conf: Configuration):
        """Take a conf object (from the Configuration class) to run a deep_classifier engine.

        :param conf: a conf object
        """
        super().__init__(conf)

    @timeit("compute the predictions scores")
    def score_prediction(self):
        """Score the predictions of the models.

        :return: None
        """
        if self._conf.data.validation_score_activate:
            logging.info("Computing the prediction scores...")
            scorer = ClassifierScorer(self.conf, self.models_dict, self.processor)
            scorer.score()



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This modules implements multiple classes to handle operations related to tabular data for deep engines, i.e.
deep_classifier and deep_regressor engines. This module communicates with other data worker modules to load, validate,
and process the data and create datasets, etc."""

import logging

from torch.utils.data import DataLoader

from learner.configuration.configuration import Configuration
from learner.data_worker.data_processor import DataProcessor
from learner.data_worker.data_sampler import DataSampler
from learner.data_worker.output_handler import OutputHandler
from learner.data_worker.data_set import (LearnerTrainTabularDataset, LearnerValidationTabularDataset,
                                          LearnerTestTabularDataset)
from learner.feature_engineering.feature_engineering import FeatureEngineering
from learner.validator.data_validator import DataValidator
from learner.outlier_manager.outlier_manager import OutlierManager
from learner.utilities.timer import timeit
from learner.validator.input_validator import remove_subset_list
from learner.analysis.analysis import Analysis


class TrainDataManager:
    """This class manages the training data for deep engines. The main method in this class is `get_data_loader` method.
    Unlike the class in data_manager module, we need to explicitly call that method to get the data_loader."""

    def __init__(self, conf: Configuration, mdl=None):
        """Initialize a TrainDataManager object using a conf object and an item in models_dict.

        :param conf: a conf object
        :param mdl: an item of models_dict. We use this to save the model classes.
        """
        self._conf = conf
        self._mdl = mdl
        self.data = None
        self.processor = None
        self.feature_engineering = None
        self.sample = None

    @property
    def conf(self):
        return self._conf

    @property
    def mdl(self):
        return self._mdl

    @timeit("load the training data")
    def get_data(self):
        """Load the training data. Then, do all the validation, processing, and feature engineering. The method
        `get_data_loader` calls this method. Please note that we import the learner DataLoader class here to avoid name
        conflicts with the PyTorch class.

        :return: None
        """
        logging.info("Loading the train data...")
        from learner.data_worker.data_loader import DataLoader
        loader = DataLoader(self._conf)
        self.data = loader.load_train_data()
        self.save_data()
        self.validate_data()
        self.analyze_data()
        self.sample_data()
        self.handle_outlier()
        self.process_data()
        self.handle_feature_engineering()

    @timeit("save the data")
    def save_data(self):
        """If requested by user, save the data obtained using the query into a csv file.

        :return: None
        """
        if self._conf.data.train_query_activate and self._conf.data.train_query_save_to_file:
            format_ = self._conf.data.train_format
            logging.info(f"Saving the train data obtained from the query into a {format_} file")
            output = OutputHandler(self._conf)
            filename = f"train_data_from_query{self.conf.sep_timetag}.{format_}"
            output.save_file(self.conf.workspace.data_directory,
                             filename,
                             self.data,
                             add_timetag=False,
                             format=format_,
                             sep=self.conf.data.train_delimiter)
            logging.info("Successfully saved the train data into the data directory")

    @timeit("filter the outliers")
    def handle_outlier(self):
        """Instantiate an outlier manager object and make the call to handle_outlier to get the updated data.

        :return: None
        """
        outlier_manager = OutlierManager(self._conf, self.data)
        self.data = outlier_manager.handle_outlier()

    @timeit("validate the training data")
    def validate_data(self):
        """Validate the data

        :return: None
        """
        if self._conf.validation.validate:
            validator = DataValidator(self._conf, self.data, data_type="train")
            validator.validate_data()

    @timeit("analyze the training data")
    def analyze_data(self):
        """Analyze the data

        :return: None
        """
        analysis = Analysis(self._conf)
        analysis.analyze_data(self.data)

    @timeit("sample the training data")
    def sample_data(self):
        """Sample the training data if necessary. Note that we may not need to sample the data in many situations. In
        that case, the driver method (sample_data) would make the decision.

        :return: the updated training dataset
        """
        self.sample = DataSampler(self.conf, self.data)
        self.data = self.sample.sample_data()

    @timeit("process the training data")
    def process_data(self):
        """Process the data. Data processing is carried out based on the user's instructions in the configuration file.

        :return: None
        """
        self.processor = DataProcessor(self._conf)
        self._conf, self.data = self.processor.process_data(self.data)

    def handle_feature_engineering(self):
        """Instantiate an object of the FeatureEngineering class and pass the data to the handle_feature_engineering
        method to do any requested feature engineering.

        :return: None
        """
        self.feature_engineering = FeatureEngineering(self._conf)
        self.data = self.feature_engineering.handle_feature_engineering(self.data)

    def get_dataset(self, df):
        """Use LearnerTrainTabularDataset class and the dataframe to obtain a dataset object. Please see the
        documentation of the dataset class to read more.

        :params: a pandas dataframe that contains information about the path to the images and their corresponding label
        :return: a LearnerTrainTabularDataset object
        """
        dataset = LearnerTrainTabularDataset(df, conf=self.conf, mdl=self.mdl)
        return dataset

    def get_data_loader(self):
        """The main method for getting the train data loader. This method calls other methods in this class to obtain
        all the necessary information for creating a DataLoader object. This includes getting the dataframe and
        the dataset. We also update the mdl dictionary but populating a list of training features.

        :return: a PyTorch DataLoader object
        """
        logging.info("Getting the train data loader...")
        self.get_data()
        # save the train_features
        self.mdl["train_features"] = remove_subset_list(list(self.data.drop(self._conf.column.drop_from_train, axis=1).columns),
                                                        [self._conf.column.target_col])
        dataset = self.get_dataset(self.data)

        batch_size = self.mdl["params"]["batch_size"] if "batch_size" in self.mdl["params"] and self.mdl["params"]["batch_size"] else self.conf.data.train_batch_size
        return DataLoader(dataset, batch_size=batch_size, shuffle=True)


class ValidationDataManager:
    """This class manages the validation data for deep engines. The main method in this class is the `get_data_loader`
    method.  Unlike the class in data_manager module, we need to explicitly call that method to get the data_loader."""

    def __init__(self, conf, mdl, processor=None, feature_engineering=None):
        """Initialize a ValidationDataManager object using a conf object, an item in models_dict, as well as the
        processor and feature_engineering objects.

        :param conf: a conf object
        :param processor: a DataProcessor object
        :param feature_engineering: a FeatureEngineering object
        """
        self._conf = conf
        self._mdl = mdl
        self.data = None
        self.processor = processor
        self.feature_engineering = feature_engineering

    @property
    def conf(self):
        return self._conf

    @property
    def mdl(self):
        return self._mdl

    @timeit("load the validation data")
    def get_data(self):
        """Load the training data. Then, do all the validation, processing, and feature engineering. The method
        `get_data_loader` calls this method. Please note that we import the learner DataLoader class here to avoid name
        conflicts with the PyTorch class.

        :return: None
        """
        logging.info("Loading the validation data...")
        from learner.data_worker.data_loader import DataLoader
        loader = DataLoader(self._conf)
        self.data = loader.load_validation_data()
        self.save_data()
        self.validate_data()
        self.handle_outlier()
        self.process_data()
        self.handle_feature_engineering()

    @timeit("save the data")
    def save_data(self):
        """If requested by user, save the data obtained using the query into a csv file.

        :return: None
        """
        if self._conf.data.validation_query_activate and self._conf.data.validation_query_save_to_file:
            format_ = self._conf.data.validation_format
            logging.info(f"Saving the validation data obtained from the query into a {format_} file")
            output = OutputHandler(self._conf)
            filename = f"validation_data_from_query{self.conf.sep_timetag}.{format_}"
            output.save_file(self.conf.workspace.data_directory,
                             filename,
                             self.data,
                             add_timetag=False,
                             format=format_,
                             sep=self.conf.data.train_delimiter)
            logging.info("Successfully saved the validation data into the data directory")

    @timeit("filter the outliers")
    def handle_outlier(self):
        """Instantiate an outlier manager object and make the call to handle_outlier to get the updated data.

        :return: None
        """
        outlier_manager = OutlierManager(self._conf, self.data)
        self.data = outlier_manager.handle_outlier()

    @timeit("validate the validation data")
    def validate_data(self):
        """Validate the data

        :return: None
        """
        if self._conf.validation.validate:
            validator = DataValidator(self._conf, self.data, data_type="validation")
            validator.validate_data()

    @timeit("process the validation data")
    def process_data(self):
        """Process the data. Data processing is carried out based on the user's instructions in the configuration file.

        :return: None
        """
        self._conf, self.data = self.processor.process_data(self.data)

    def handle_feature_engineering(self):
        """Instantiate an object of the FeatureEngineering class and pass the data to the handle_feature_engineering
        method to do any requested feature engineering.

        :return: None
        """
        self.feature_engineering = FeatureEngineering(self._conf)
        self.data = self.feature_engineering.handle_feature_engineering(self.data)

    def get_dataset(self, df):
        """Use LearnerValidationTabularDataset class and the dataframe to obtain a dataset object. Please see the
        documentation of the dataset class to read more.

        :params: a pandas dataframe that contains information about the path to the images and their corresponding label
        :return: a LearnerValidationTabularDataset object
        """
        dataset = LearnerValidationTabularDataset(df, conf=self.conf, mdl=self.mdl)
        return dataset

    def get_data_loader(self):
        """The main method for getting the validation data loader. This method calls other methods in this class to
        obtain all the necessary information for creating a DataLoader object. This includes getting the dataframe and
        the dataset.

        :return: a PyTorch DataLoader object
        """
        logging.info("Getting the validation data loader...")
        self.get_data()
        dataset = self.get_dataset(self.data)

        return DataLoader(dataset, batch_size=self.conf.data.validation_batch_size, shuffle=False)


class TestDataManager:
    """This class manages the test data for deep engines. Because we need to be able handle dataset of any sizes, this
    class works slightly different from the train and validation classes. Here, we get data reader and chunk of data
    first. We then pass that chunk of data to get a data loader. We then iterate through the reader object so that we
    load a limited amount of data at a time."""

    # setting this so that pytest would not try to run this as a test
    __test__ = False

    def __init__(self, conf, processor=None, feature_engineering=None):
        """Initialize a TestDataManager object using a conf object as well as the processor and feature_engineering
        objects.

        :param conf: a conf object
        :param processor: a DataProcessor object
        :param feature_engineering: a FeatureEngineering object
        """
        self._conf = conf
        self.processor = processor
        self.feature_engineering = feature_engineering
        self.save_data()

    @property
    def conf(self):
        return self._conf

    def save_data(self):
        """In some situations, we need to load the entire test data in memory and the save it into disk. These
        situations include using query and segmentation or using query and setting save_to_file to true. In those cases,
        use data loader to get the data and save it to a csv file.

        :return: None
        """
        # if requested or segmenter's been activated, we need to get all the test data and save it
        if (self._conf.data.test_query_activate and self._conf.data.test_query_save_to_file) or \
           (self._conf.data.test_query_activate and self.conf.segmenter.activate):
            format_ = self._conf.data.test_format
            logging.info(f"Getting the test data using the query and saving it into a {format_} file")
            from learner.data_worker.data_loader import DataLoader
            loader = DataLoader(self._conf)
            data = loader.load_test_from_db()
            data = data[self.conf.column.use_cols] if self.conf.column.use_cols else data
            output = OutputHandler(self._conf)
            filename = f"test_data_from_query{self.conf.sep_timetag}.csv"
            output.save_file(self.conf.workspace.data_directory,
                             filename,
                             data,
                             add_timetag=False,
                             format=format_,
                             sep=self.conf.data.test_delimiter)
            logging.info("Successfully saved the test data into the data directory. Learner will use this file for "
                         "making predictions moving forward")
            self.conf.data.test_location = self.conf.workspace.data_directory + filename

    def get_reader(self):
        """Use load_test_data to get a TextFileReader object

        :return: TextFileReader object
        """
        logging.info("Getting a reader to process the test data...")
        from learner.data_worker.data_loader import DataLoader
        loader = DataLoader(self._conf)
        reader = loader.load_test_data()
        # when using csv files, reader is an iterable otherwise is the full df. We return an iterable (a list in this
        # case) so we can treat all formats the same
        if self._conf.data.test_format == "csv":
            return reader
        else:
            return [reader]

    def get_reader_data(self, chunk):
        """Validate, process, and return the processed dataset for a chunk.

        :param chunk: a pandas dataframe
        :return: a processed pandas dataframe
        """
        # make sure we only keep use_cols if we are not reading from a file
        if not self._conf.data.test_location:
            chunk = chunk[remove_subset_list(self._conf.column.use_cols, [self._conf.column.target_col])]

        self.validate_data(chunk)
        _, chunk = self.process_data(chunk)
        chunk = self.handle_feature_engineering(chunk)
        return chunk

    def validate_data(self, data):
        """Validate the data.

        :param data: a pandas dataframe
        :return: None
        """
        if self._conf.validation.validate:
            validator = DataValidator(self._conf, data, data_type="test")
            validator.validate_data()

    def process_data(self, data):
        """Process the data. Data processing is carried out based on user input in the config file.

        :param data: a pandas dataframe to be processed
        :return: a processed pandas dataframe
        """
        # we'd probably never need to check this
        if self.processor is None:
            self.processor = DataProcessor(self._conf)
        _, data = self.processor.process_data(data)
        return _, data

    def handle_feature_engineering(self, chunk):
        """Get a chunk and run it through various feature_engineering processes.

        :param chunk: a pandas dataframe loaded in memory
        :return: the updated dataframe after feature engineering
        """
        if self.feature_engineering is None:
            self.feature_engineering = FeatureEngineering(self._conf)
        chunk = self.feature_engineering.handle_feature_engineering(chunk)
        return chunk

    def get_dataset(self, chunk, mdl):
        """Use LearnerTestTabularDataset class, the dataframe, and an item of models_dict to obtain a dataset object.
        Please see the documentation of the dataset class to read more.

        :param chunk: a pandas dataframe that contains the loaded data
        :param mdl: an item (value) in models_dict.
        :return: a LearnerTestTabularDataset object
        """
        dataset = LearnerTestTabularDataset(chunk, conf=self.conf, mdl=mdl)
        return dataset

    def get_data_loader(self, chunk, mdl):
        """The method for getting the test data loader for a chunk of data. Unlike the train and validation data, this
        is not the main class. Here, we need to provide that loaded data as an argument. The `get_reader_data` method
        handles loading and processing the data.

        :param chunk: a pandas dataframe that contains the loaded data
        :param mdl: an item (value) in models_dict.
        :return: a PyTorch DataLoader object
        """
        logging.info("Getting the test data loader...")
        dataset = self.get_dataset(chunk, mdl)
        return DataLoader(dataset, batch_size=self.conf.data.test_batch_size, shuffle=False)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements classes and functions for loading image data. The main class is ImageDataLoader. This class
uses other class and functions to identify and load image data. Learner uses a consistent approach for treating image
data. If the image data are organized in folders, Learner creates a dataframe to identify the path to the images and
their corresponding labels, if relevant. If a file is provided, Learner will read that file."""

import sys
import os

from torchvision import datasets
import pandas as pd

from learner.data_worker.data_loader import DataLoader


class ImageDataLoader(DataLoader):
    """A class to create a dataframe containing image path (and labels if relevant) from image data. This class
    implements three main methods to create dataframes for train, validation, and test data. Depending on the input
    type, i.e. "file" or "folder", the relevant classes and methods are used.
    """
    def __init__(self, conf, mdl):
        """Initialize a ImageDataLoader object using a conf object and an item in models_dict.

        :param conf: a conf object
        :param mdl: an item of models_dict. We use this to save the model classes.
        """
        super(ImageDataLoader, self).__init__(conf)
        self._conf = conf
        self._mdl = mdl

    @property
    def conf(self):
        return self._conf

    @property
    def mdl(self):
        return self._mdl

    def get_train_df(self):
        """Construct the dataframe corresponding to the train image data. If the input type is "folder", we use the
        DataFrameFromImageFolder class to get the dataframe. If the input type is "file", we use the parent method to
        obtain the dataframe.

        :return: a pandas dataframe that contains the path to the images and the corresponding labels.
        """
        if self._conf.data.train_input_type == "folder":
            df = DataFrameFromImageFolder(conf=self._conf, root=self._conf.data.train_location, data_type="train", mdl=self._mdl).get_df()
        else:
            df = self.load_train_from_file()
            self._mdl["classes"] = sorted(list(df[self._conf.column.target_col].unique()))
            from learner.data_worker.data_processor import LearnerLabelEncoder
            LearnerLabelEncoder().fit_transform(df, cols=[self._conf.column.target_col])
        return df

    def get_validation_df(self):
        """Construct the dataframe corresponding to the validation image data. If the input type is "folder", we use the
        DataFrameFromImageFolder class to get the dataframe. If the input type is "file", we use the parent method to
        obtain the dataframe.

        :return: a pandas dataframe that contains the path to the images and the corresponding labels.
        """
        if self._conf.data.validation_input_type == "folder":
            df = DataFrameFromImageFolder(conf=self._conf, root=self._conf.data.validation_location, data_type="validation", mdl=self._mdl).get_df()
        else:
            df = self.load_validation_from_file()
            self._mdl["classes"] = sorted(list(df[self._conf.column.target_col].unique()))
            from learner.data_worker.data_processor import LearnerLabelEncoder
            LearnerLabelEncoder().fit_transform(df, cols=[self._conf.column.target_col])
        return df

    def get_test_df(self):
        """Construct the dataframe corresponding to the test image data. If the input type is "folder", we use the
        DataFrameFromImageFolder class to get the dataframe. If the input type is "file", we use the parent method to
        obtain the dataframe.

        :return: a pandas dataframe that contains the path to the images. Unlike other methods, the test data does not have labels.
        """
        if self._conf.data.test_input_type == "folder":
            df = DataFrameFromImageFolder(conf=self._conf, root=self._conf.data.test_location, data_type="test").get_df()
        else:
            df = self.load_test_from_file(chunksize=None)
        return df


class DataFrameFromImageFolder:
    """A class to construct a pandas dataframe for an image folder. The assumption is that the images are arranged in
    the following patterns.

    /cat/1.jpg
    /cat/2.jpg

    /dog/1.jpg
    /dog/2.jpg

    This classes uses pytorch's functionality whenever possible. For example, supported image extensions comes from
    pytorch. We also use ImageFolder class of pytorch to get the dataframe for train and validation data.
    """
    def __init__(self, conf, root, data_type, mdl=None):
        """Initialize a DataFrameFromImageFolder object.

        :param conf: a conf object. It's mainly used for getting the column names.
        :param root: the path to the directory where all the subfolders are located.
        :param data_type: the data type, it should be "train", "test", or "validation".
        :param mdl: an item of models_dict dictionary.
        """
        self._conf = conf
        self._root = root
        self._data_type = data_type
        self._mdl = mdl
        self.extensions = datasets.folder.IMG_EXTENSIONS

    @property
    def conf(self):
        return self._conf

    @property
    def root(self):
        return self._root

    @property
    def data_type(self):
        return self._data_type

    @property
    def mdl(self):
        return self._mdl

    def get_df(self):
        """Depending on the data type, call the relevant method to obtain the dataframes.

        :return: The pandas dataframe that are returned by other methods.
        """
        if self.data_type == "test":
            return self.get_test_df()
        else:
            return self.get_train_validation_df()

    def get_test_df(self):
        """Leverage the function get_image_list_from_directory to get a list of images in the directory. Then, create
        a dataframe using that list.

        :return: a pandas dataframe with a column containing the path to all the images.
        """
        samples = get_image_list_from_directory(self.root, self.extensions)
        if len(samples) == 0:
            sys.exit(f"Found 0 files in {self.root}. Supported extensions are: {self.extensions}")
        return pd.DataFrame(samples, columns=[self._conf.column.path_col])

    def get_train_validation_df(self):
        """Use ImageFolder class from pytorch to get the list of images and labels. Then use that list to create the
        dataframe.

        :return: a pandas dataframe with three columns: the path column, the target column, and the classes (from the folder names).
        """
        validation_data = datasets.ImageFolder(root=self.root)
        self.mdl["classes"] = validation_data.classes
        df = pd.DataFrame(validation_data.samples, columns=[self._conf.column.path_col, self._conf.column.target_col])
        df["classes"] = df[self._conf.column.target_col].map({index: class_ for class_, index in validation_data.class_to_idx.items()})
        return df


def get_image_list_from_directory(directory, extensions):
    """Accept the path to a directory and the list of extensions (this usually comes from PyTorch) and return a list of
    paths to the images.

    :param directory: the path to the directory to look for the images
    :param extensions: the list of acceptable extensions
    :return: a list of paths to the images in the directory
    """
    def is_valid_file(x):
        return datasets.folder.has_file_allowed_extension(x, extensions)

    instances = []
    directory = os.path.expanduser(directory)

    for root, _, fnames in sorted(os.walk(directory, followlinks=True)):
        for fname in sorted(fnames):
            path = os.path.join(root, fname)
            if is_valid_file(path):
                instances.append(path)
    return instances




# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
from math import e
import warnings
import logging

from learner.validator.input_validator import validate_intersection_cols, validate_subset_list
from learner.data_worker.data_loader import get_value
from learner.configuration.column import ColumnConfiguration
from learner.configuration.sample import SampleConfiguration
from learner.configuration.supported_items import DATE_ITEMS, SUPPORTED_ENGINES_FOR_TARGET_LOG_TRANSFORM, SUPPORTED_STEMMERS
from learner.validator.model_validator import DefaultValidatingDraft7Validator
from learner.configuration.defaults import IMAGE_TRANSFORM_PARAMS


class ProcessConfiguration:
    """Parse all input variables related to data processing."""

    def __init__(self, json_config, data, column: ColumnConfiguration, sample: SampleConfiguration, engine):
        self._json_config = json_config
        self._data = data
        self._column = column
        self._sample = sample
        self._engine = engine

        self.fillnan_activate = self.get_fillnan_activate()
        self.fillnan_mean_cols = get_value(self._json_config, [], "process", "fillnan_params", "mean_cols")
        self.fillnan_median_cols = get_value(self._json_config, [], "process", "fillnan_params", "median_cols")
        self.fillnan_mode_cols = get_value(self._json_config, [], "process", "fillnan_params", "mode_cols")
        self.fillnan_value_cols = self.get_fillnan_value_cols()
        self.fillnan_value = get_value(self._json_config, 0, "process", "fillnan_params", "value")

        self.dummies_activate = self.get_dummies_activate()
        self.dummies_cols = self.get_dummies_cols()

        self.to_numeric_activate = self.get_to_numeric_activate()
        self.to_numeric_cols = self.get_to_numeric_cols()

        self.label_encoding_activate = self.get_label_encoding_activate()
        self.label_encoding_cols = self.get_label_encoding_cols()
        self.label_encode_target = self.get_label_encode_target()

        self.tuplize_activate = get_value(self._json_config, False, "process", "tuplize_params", "activate")
        self.tuplize_cols = self.get_tuplize_cols()

        self.date_cols_activate = self._column.date_cols_activate
        self.date_cols_params = self.get_date_cols_params()

        self.standard_scaler_activate = get_value(self._json_config, False, "process", "standard_scaler_params", "activate")
        self.standard_scaler_cols = self.get_standard_scaler_cols()

        self.min_max_scaler_activate = get_value(self._json_config, False, "process", "min_max_scaler_params", "activate")
        self.min_max_scaler_cols = self.get_min_max_scaler_cols()

        self.memory_optimization = get_value(self._json_config, False, "process", "memory_optimization")

        self.log_transform_target_activate = self.get_log_transform_target_activate()
        self.log_transform_target_base = get_value(self._json_config, e, "process", "log_transform_target", "base")
        self.log_transform_target_shift = get_value(self._json_config, 0, "process", "log_transform_target", "shift")
        self.log_transform_target_params = self.get_log_transform_target_params()
        self.log_transform_target_predict_actuals = self.get_log_transform_target_predict_actuals()
        self.log_transform_target_score_actuals = self.get_log_transform_target_score_actuals()

        self.lemmatize_activate = get_value(self._json_config, False, "process", "lemmatize_params", "activate")
        self.lemmatize_cols_params = self.get_lemmatize_cols_params()

        self.stem_activate = get_value(self._json_config, False, "process", "stem_params", "activate")
        self.stem_cols_params = self.get_stem_cols_params()

        self.count_vectorize_activate = get_value(self._json_config, False, "process", "count_vectorize_params", "activate")
        self.count_vectorize_cols_params = self.get_count_vectorize_cols_params()

        self.tfidf_activate = get_value(self._json_config, False, "process", "tfidf_params", "activate")
        self.tfidf_cols_params = self.get_tfidf_cols_params()

        self.image_transform_params = self.get_image_transform_params()

    def get_fillnan_activate(self):
        try:
            activate = self._json_config["process"]["fillnan_params"]["activate"]
            if not activate:
                warnings.warn("fillnan_activate was set to false, this may lead to errors...", UserWarning)
            return activate
        except KeyError:
            warnings.warn("Using the default value, false, for fillnan_activate, this may lead to errors...", UserWarning)
            return False

    def get_fillnan_value_cols(self):
        value_cols = {}
        try:
            value_cols = self._json_config["process"]["fillnan_params"]["value_cols"]
        except KeyError:
            value_cols = {}
        finally:
            basic_operations_names = self.get_basic_operations_names()
            log_transform_names = self.get_log_transform_names()
            self._column.valid_cols.extend(basic_operations_names + log_transform_names)
            if self.fillnan_activate:
                self._validate_update_fillnan_cols(value_cols)
            return value_cols

    def get_basic_operations_names(self):
        try:
            basic_operations_params = self._json_config["feature_engineering"]["basic_operations_params"]
            names = []
            for item in basic_operations_params:
                try:
                    names.append(item["name"])
                except KeyError:
                    pass
            return names
        except KeyError:
            return []

    def get_log_transform_names(self):
        try:
            log_transform_params = self._json_config["feature_engineering"]["log_transform_params"]
            names = []
            for item in log_transform_params:
                try:
                    names.append(item["name"])
                except KeyError:
                    pass
            return names
        except KeyError:
            return []

    def _validate_update_fillnan_cols(self, fillnan_cols):
        if self.fillnan_mean_cols:
            validate_intersection_cols(cols1=self.fillnan_mean_cols, cols2=fillnan_cols.keys(),
                                       cols1_name="mean cols", cols2_name="value cols")
            if self.fillnan_median_cols:
                validate_intersection_cols(cols1=self.fillnan_mean_cols, cols2=self.fillnan_median_cols,
                                           cols1_name="mean cols", cols2_name="median cols")
            if self.fillnan_mode_cols:
                validate_intersection_cols(cols1=self.fillnan_mean_cols, cols2=self.fillnan_mode_cols,
                                           cols1_name="mean cols", cols2_name="mode cols")

        if self.fillnan_median_cols:
            validate_intersection_cols(cols1=self.fillnan_median_cols, cols2=fillnan_cols.keys(),
                                       cols1_name="median cols", cols2_name="value cols")
            if self.fillnan_mean_cols:
                validate_intersection_cols(cols1=self.fillnan_median_cols, cols2=self.fillnan_mean_cols,
                                           cols1_name="median cols", cols2_name="mean cols")
            if self.fillnan_mode_cols:
                validate_intersection_cols(cols1=self.fillnan_median_cols, cols2=self.fillnan_mode_cols,
                                           cols1_name="median cols", cols2_name="mode cols")

        if self.fillnan_mode_cols:
            validate_intersection_cols(cols1=self.fillnan_mode_cols, cols2=fillnan_cols.keys(),
                                       cols1_name="mode cols", cols2_name="value cols")
            if self.fillnan_mean_cols:
                validate_intersection_cols(cols1=self.fillnan_mode_cols, cols2=self.fillnan_mean_cols,
                                           cols1_name="mode cols", cols2_name="mean cols")
            if self.fillnan_median_cols:
                validate_intersection_cols(cols1=self.fillnan_mode_cols, cols2=self.fillnan_median_cols,
                                           cols1_name="mode cols", cols2_name="median cols")

    def get_dummies_activate(self):
        try:
            activate = self._json_config["process"]["dummies_params"]["activate"]
            if not activate:
                warnings.warn("dummies_activate was set to false, this may lead to errors...", UserWarning)
            return activate
        except KeyError:
            warnings.warn("Using the default value, false, for dummies_activate, this may lead to errors...",
                          UserWarning)
            return False

    def get_dummies_cols(self):
        cat_cols = set()
        try:
            cat_cols = set(self._json_config["process"]["dummies_params"]["cols"])
            # for deep learning engines, on-hot encoding does not make sense because embedding basically does it
            if self._engine in ["DeepClassifier", "DeepRegressor"] and self.dummies_activate:
                logging.critical("You are using a deep learning engine. All categorical columns should be defined in "
                                 "label_encoding_params section for embedding. Please update the configuration file "
                                 "and try again. Exiting...")
                sys.exit(1)
        finally:
            if not cat_cols and self._data.meta_data:
                logging.info("No dummies cols was defined in the configuration file. Trying to get them from the "
                             "meta_data file...")
                for col in self._data.meta_data["column"]:
                    if col in self._column.use_cols:
                        if self._data.meta_data["column"][col]["type"] == "dummy":
                            cat_cols.add(col)
            return list(cat_cols)

    def get_to_numeric_activate(self):
        try:
            activate = self._json_config["process"]["to_numeric_params"]["activate"]
            if not activate:
                warnings.warn("to_numeric_activate was set to false, this may lead to errors...", UserWarning)
            return activate
        except KeyError:
            warnings.warn("Using the default value, false, for to_numeric_activate, this may lead to errors...",
                          UserWarning)
            return False

    def get_to_numeric_cols(self):
        try:
            num_cols = self._json_config["process"]["to_numeric_params"]["cols"]
            return num_cols
        except KeyError:
            return []

    def get_label_encoding_activate(self):
        try:
            activate = self._json_config["process"]["label_encoding_params"]["activate"]
            if not activate:
                warnings.warn("label_encoding_activate was set to false, this may lead to errors...", UserWarning)
            return activate
        except KeyError:
            warnings.warn("Using the default value, false, for label_encoding_activate, this may lead to errors...",
                          UserWarning)
            return False

    def get_label_encoding_cols(self):
        try:
            label_encode_cols = self._json_config["process"]["label_encoding_params"]["cols"]
            if self.label_encoding_activate and self.dummies_activate:
                validate_intersection_cols(cols1=self.dummies_cols, cols2=label_encode_cols,
                                           cols1_name='dummies cols', cols2_name='label_encoding cols')
            return label_encode_cols
        except KeyError:
            return []

    def get_label_encode_target(self):
        if self._engine in ["Classifier", "DeepClassifier"] and self.label_encoding_activate and self._column.target_col in self.label_encoding_cols:
            return True
        return False

    def get_tuplize_cols(self):
        try:
            tuplize_cols = self._json_config["process"]["tuplize_params"]["cols"]
            return tuplize_cols
        except KeyError:
            return None

    def get_date_cols_params(self):
        try:
            date_cols = self._json_config["process"]["date_cols_params"]["cols_params"]
            if self.date_cols_activate and date_cols:
                for col, date_item in date_cols.items():
                    # validate requested date items
                    validate_subset_list(DATE_ITEMS,
                                         "acceptable date items",
                                         subset_list=date_item,
                                         subset_name="requested items")

                # we know that date columns cannot be reached to the training
                diff = set(date_cols) - set(self._column.drop_from_train)
                if diff:
                    warnings.warn("The columns(s) {0} exist in date cols but not in drop_from_train. Updating"
                                  "drop_from_train to include date cols".format(diff), UserWarning)
                    self._column.drop_from_train = list(set(self._column.drop_from_train).union(set(date_cols)))

            return date_cols
        except KeyError:
            return False

    def get_standard_scaler_cols(self):
        try:
            standard_scaler_cols = self._json_config["process"]["standard_scaler_params"]["cols"]
            if self.standard_scaler_activate:
                # we need to add these columns to numeric_cols
                logging.info("Adding standard_scaler columns to numeric_cols if they are not already included")
                self.to_numeric_cols = list(set(self.to_numeric_cols).union(set(standard_scaler_cols)))
                self.to_numeric_activate = True
            return standard_scaler_cols
        except KeyError:
            return None

    def get_min_max_scaler_cols(self):
        try:
            min_max_scaler_cols = self._json_config["process"]["min_max_scaler_params"]["cols"]
            if self.min_max_scaler_activate:
                # we need to add these columns to numeric_cols
                logging.info("Adding min_max_scaler columns to numeric_cols if they are not already included")
                self.to_numeric_cols = list(set(self.to_numeric_cols).union(set(min_max_scaler_cols)))
                self.to_numeric_activate = True
            return min_max_scaler_cols
        except KeyError:
            return None

    def get_log_transform_target_activate(self):
        try:
            activate = self._json_config["process"]["log_transform_target"]["activate"]
            if activate:
                validate_subset_list(parent_list=SUPPORTED_ENGINES_FOR_TARGET_LOG_TRANSFORM,
                                     parent_name="acceptable engines for target log transform",
                                     subset_list=[self._engine],
                                     subset_name="defined engine")
            return activate
        except KeyError:
            return False

    def get_log_transform_target_params(self):
        if self.log_transform_target_activate:
            return [{"activate": True,
                     "base": self.log_transform_target_base,
                     "col": self._column.target_col,
                     "name": self._column.target_col,
                     "shift": self.log_transform_target_shift}]
        return []

    def get_log_transform_target_predict_actuals(self):
        if not self.log_transform_target_activate:
            return False
        try:
            predict_actuals = self._json_config["process"]["log_transform_target"]["predict_actuals"]
            if predict_actuals and not self._data.test_prediction_activate:
                logging.critical("prediction_params in test_params was not activated but predict_actuals in "
                                 "log_transform_target was set to true. Learner doesn't know how to proceed. "
                                 "Please update your configuration file and try again. Exiting...")
                sys.exit(1)
            return predict_actuals
        except KeyError:
            return False

    def get_log_transform_target_score_actuals(self):
        if not self.log_transform_target_activate:
            return False
        try:
            score_actuals = self._json_config["process"]["log_transform_target"]["score_actuals"]
            if score_actuals and not self._data.validation_score_activate:
                logging.critical("scoring_params in validation_params was not activated but score_actuals in "
                                 "log_transform_target was set to true. Learner doesn't know how to proceed. "
                                 "Please update your configuration file and try again. Exiting...")
                sys.exit(1)
            return score_actuals
        except KeyError:
            return False

    def get_lemmatize_cols_params(self):
        lemmatize_default_params = {"pos": ["v"]}
        try:
            cols_params = self._json_config["process"]["lemmatize_params"]["cols_params"]
            self._validate_lemmatize_params(cols_params, lemmatize_default_params)
            return cols_params
        except KeyError:
            if self.lemmatize_activate:
                logging.critical("lemmatize_params was activated but cols_params was not defined. You can either "
                                 "deactivate this feature or define cols_params. Exiting...")
                sys.exit(1)
            return []

    def _validate_lemmatize_params(self, cols_params, default_params):
        for col_param in cols_params:
            # make sure all the keys are valid
            validate_subset_list(parent_list=list(default_params.keys()) + ["name"],
                                 parent_name="acceptable fields in lemmatize cols_params",
                                 subset_list=col_param.keys(),
                                 subset_name="provided fields")

            if "name" not in col_param:
                logging.critical(f"each dictionary in cols_params in the lemmatize_params of the process section "
                                 f"should define the column name using the 'name' field. Exiting...")
                sys.exit(1)
            # make sure pos is valid
            if "pos" in col_param:
                validate_subset_list(parent_list=['a', 'r', 'n', 'v'],
                                     parent_name=f"valid pos for lemmatizer",
                                     subset_list=col_param["pos"],
                                     subset_name=f"defined pos for lemmatizer")

            # if the parameter is defined use it, otherwise use the default value
            for param, default_value in default_params.items():
                col_param[param] = col_param.get(param, default_value)

    def get_stem_cols_params(self):
        stem_default_params = {"type": "PosterStemmer",
                               "options": {}}
        try:
            cols_params = self._json_config["process"]["stem_params"]["cols_params"]
            self._validate_stem_params(cols_params, stem_default_params)
            return cols_params
        except KeyError:
            if self.stem_activate:
                logging.critical("stem_params was activated but cols_params was not defined. You can either "
                                 "deactivate this feature or define cols_params. Exiting...")
                sys.exit(1)
            return []

    def _validate_stem_params(self, cols_params, default_params):
        for col_param in cols_params:
            # make sure all the keys are valid
            validate_subset_list(parent_list=list(default_params.keys()) + ["name"],
                                 parent_name="acceptable fields in stem cols_params",
                                 subset_list=col_param.keys(),
                                 subset_name="provided fields")

            if "name" not in col_param:
                logging.critical(f"each dictionary in cols_params in the stem_params of the process section "
                                 f"should define the column name using the 'name' field. Exiting...")
                sys.exit(1)
            # make sure pos is valid
            if "type" in col_param:
                validate_subset_list(parent_list=SUPPORTED_STEMMERS,
                                     parent_name=f"valid type for stemmer",
                                     subset_list=[col_param["type"]],
                                     subset_name=f"defined pos for stemmer")

            # if the parameter is defined use it, otherwise use the default value
            for param, default_value in default_params.items():
                col_param[param] = col_param.get(param, default_value)

    def get_count_vectorize_cols_params(self):
        count_vectorize_default_params = {
            "strip_accents": None,
            "lowercase": True,
            "stop_words": None,
            "token_pattern": r"(?u)\b\w\w+\b",
            "ngram_range": [1, 1],
            "analyzer": "word",
            "max_df": 1.0,
            "min_df": 1,
            "max_features": None,
            "binary": False
        }
        try:
            cols_params = self._json_config["process"]["count_vectorize_params"]["cols_params"]
            self._validate_count_vectorize_and_tfidf_params(cols_params, count_vectorize_default_params, type_="count_vectorize")
            return cols_params
        except KeyError:
            if self.count_vectorize_activate:
                logging.critical("count_vectorize_params was activated but cols_params was not defined. You can either "
                                 "deactivate this feature or define cols_params. Exiting...")
                sys.exit(1)
            return []

    def get_tfidf_cols_params(self):
        tfidf_default_params = {
            "strip_accents": None,
            "lowercase": True,
            "stop_words": None,
            "token_pattern": r"(?u)\b\w\w+\b",
            "ngram_range": [1, 1],
            "analyzer": "word",
            "max_df": 1.0,
            "min_df": 1,
            "max_features": None,
            "binary": False,
            "norm": "l2",
            "use_idf": True,
            "smooth_idf": True,
            "sublinear_tf": False
        }
        try:
            cols_params = self._json_config["process"]["tfidf_params"]["cols_params"]
            self._validate_count_vectorize_and_tfidf_params(cols_params, tfidf_default_params, type_="tfidf")
            return cols_params
        except KeyError:
            if self.tfidf_activate:
                logging.critical("tfidf_params was activated but cols_params was not defined. You can either "
                                 "deactivate this feature or define cols_params. Exiting...")
                sys.exit(1)
            return []

    @staticmethod
    def _validate_count_vectorize_and_tfidf_params(count_vectorize_cols_params, default_params, type_):
        for col_param in count_vectorize_cols_params:
            if "name" not in col_param:
                logging.critical(f"each dictionary in cols_params in the {type_}_params of the process section "
                                 f"should define the column name using the 'name' field. Exiting...")
                sys.exit(1)
            # make sure analyzer is valid
            if "analyzer" in col_param:
                validate_subset_list(parent_list=['word', 'char', 'char_wb'],
                                     parent_name=f"valid analyzer for {type_}",
                                     subset_list=[col_param["analyzer"]],
                                     subset_name=f"defined analyzer for {type_}")
            # make sure strip_accent is valid
            if "strip_accent" in col_param:
                validate_subset_list(parent_list=['ascii', 'unicode'],
                                     parent_name=f"valid strip_accent values for {type_}",
                                     subset_list=[col_param["strip_accent"]],
                                     subset_name=f"defined strip_accent value for {type_}")

            # make sure norm is valid. This is only relevant for tfidf but we don't need to check that
            if "norm" in col_param:
                validate_subset_list(parent_list=['l1', 'l2'],
                                     parent_name=f"valid strip_accent values for {type_}",
                                     subset_list=[col_param["norm"]],
                                     subset_name=f"defined strip_accent value for {type_}")

            # if the parameter is defined use it, otherwise use the default value
            for param, default_value in default_params.items():
                col_param[param] = col_param.get(param, default_value)

            # make sure ngram_range has only two items in it
            length = len(col_param["ngram_range"])
            if length != 2:
                logging.critical(f"ngram_range in cols_params of {type_}_params should contain two integers, "
                                 f"found  {length} integers. Please update your configuration file and try again."
                                 f"Exiting...")
                sys.exit(1)
            # make sure min_n is not larger than max_n
            if col_param["ngram_range"][0] > col_param["ngram_range"][1]:
                logging.critical(f"The first number in the ngram_range {type_}_params should be smaller than "
                                 f"the second one. Please update your configuration file and try again. Exiting...")
                sys.exit(1)

    def get_image_transform_params(self):
        try:
            params = self._json_config["process"]["image_transform_params"]
            DefaultValidatingDraft7Validator(IMAGE_TRANSFORM_PARAMS).validate(params)
            image_transform_params = []
            for param in params:
                if param["activate"] is True:
                    image_transform_params.append(param)
            return image_transform_params
        except KeyError:
            return []
